from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, Sequence, Union

import torch
from torch.utils.data._utils.collate import default_collate

from eir.data_load.data_preparation_modules.imputation import (
    impute_missing_modalities_wrapper,
)
from eir.data_load.data_preparation_modules.input_preparation_wrappers import (
    prepare_inputs_memory,
)
from eir.experiment_io.experiment_io import load_transformers
from eir.predict_modules.predict_input_setup import (
    get_input_setup_function_map_for_predict,
)
from eir.predict_modules.predict_tabular_input_setup import (
    ComputedPredictTabularInputInfo,
)
from eir.serve_modules.serve_network_utils import prepare_request_input_data_wrapper
from eir.serve_modules.serve_schemas import ComputedServeTabularInputInfo, ServeLabels
from eir.setup.input_setup import (
    al_input_objects_as_dict,
    get_input_name_config_iterator,
)
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.setup.input_setup_modules.setup_tabular import ComputedTabularInputInfo
from eir.setup.schemas import InputConfig, TabularInputDataConfig, al_input_configs
from eir.train_utils.utils import call_hooks_stage_iterable
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.serve_modules.serve_experiment_io import ServeExperiment
    from eir.train import Experiment, Hooks

logger = get_logger(name=__name__)


@dataclass(frozen=True)
class ServeBatch:
    """
    Note, we keep the pre_hook_inputs around for cases like autoregressive
    generation where we use the actual input token ids. This is simpler for e.g.
    supervised learning where can use the final processed inputs (e.g. embeddings)
    directly.
    """

    pre_hook_inputs: Dict[str, torch.Tensor]
    inputs: Dict[str, torch.Tensor]
    inputs_split: Sequence[dict[str, torch.Tensor]]
    target_labels: Dict[str, Dict[str, torch.Tensor]]
    ids: list[str]


def set_up_inputs_for_serve(
    test_inputs_configs: al_input_configs,
    hooks: Union["Hooks", None],
    output_folder: str,
) -> al_input_objects_as_dict:
    all_inputs = {}

    name_config_iter = get_input_name_config_iterator(input_configs=test_inputs_configs)
    for name, input_config in name_config_iter:
        cur_input_data_config = input_config.input_info
        setup_func = get_input_setup_function_for_serve(
            input_type=cur_input_data_config.input_type
        )
        logger.info(
            "Setting up %s inputs '%s' from %s.",
            cur_input_data_config.input_type,
            cur_input_data_config.input_name,
            cur_input_data_config.input_source,
        )
        set_up_input = setup_func(
            input_config=input_config,
            ids=(),
            output_folder=output_folder,
            hooks=hooks,
        )
        all_inputs[name] = set_up_input

    return all_inputs


def general_pre_process(
    data: Sequence, serve_experiment: "ServeExperiment"
) -> ServeBatch:
    exp = serve_experiment

    inputs_parsed = parse_request_input_data_wrapper(
        data=data,
        input_objects=deepcopy(exp.inputs),
    )

    inputs_prepared = general_pre_process_raw_inputs_wrapper(
        raw_inputs=inputs_parsed,
        experiment=exp,
    )

    inputs_final = default_collate(inputs_prepared)
    inputs_final = inputs_final

    loader_batch = (inputs_final, None, None)

    batch_prep_hook_kwargs = {"experiment": exp}
    hooks = exp.hooks
    assert hooks is not None
    state = call_hooks_stage_iterable(
        hook_iterable=hooks.step_func_hooks.base_prepare_batch,
        common_kwargs={"loader_batch": loader_batch, **batch_prep_hook_kwargs},
        state=None,
    )
    batch = state["batch"]

    batch_final = ServeBatch(
        pre_hook_inputs=inputs_final,
        inputs=batch.inputs,
        inputs_split=inputs_prepared,
        target_labels={},
        ids=[f"Serve_{i}" for i in range(len(data))],
    )

    return batch_final


def parse_request_input_data_wrapper(
    data: Sequence, input_objects: al_input_objects_as_dict
) -> Sequence[dict[str, Any]]:
    loaded_data = _load_request_data(data=data)
    parsed_data = prepare_request_input_data_wrapper(
        request_data=loaded_data,
        input_objects=input_objects,
    )
    return parsed_data


def get_input_setup_function_for_serve(input_type: str) -> Callable:
    mapping_predict = get_input_setup_function_map_for_predict()

    mapping_predict["tabular"] = _setup_tabular_input_for_serve

    return mapping_predict[input_type]


def _setup_tabular_input_for_serve(
    input_config: InputConfig,
    output_folder: Path,
    *args,
    **kwargs,
) -> ComputedServeTabularInputInfo:
    input_info = input_config.input_info
    input_type_info = input_config.input_type_info

    assert isinstance(input_type_info, TabularInputDataConfig)

    cat_columns = list(input_type_info.input_cat_columns)
    con_columns = list(input_type_info.input_con_columns)
    all_columns = cat_columns + con_columns

    transformers = load_transformers(
        transformers_to_load={input_info.input_name: all_columns},
        output_folder=str(output_folder),
    )
    serve_labels = ServeLabels(label_transformers=transformers[input_info.input_name])

    serve_tabular_info = ComputedServeTabularInputInfo(
        labels=serve_labels,
        input_config=input_config,
    )

    return serve_tabular_info


def _load_request_data(data: Sequence) -> Sequence[Dict[str, Any]]:
    input_data = data
    inputs_loaded = input_data

    return inputs_loaded


def general_pre_process_raw_inputs_wrapper(
    raw_inputs: Sequence[dict[str, Any]],
    experiment: Union["Experiment", "ServeExperiment"],
) -> Sequence[dict[str, torch.Tensor]]:

    all_preprocessed = []

    for raw_input in raw_inputs:
        preprocessed = general_pre_process_raw_inputs(
            raw_inputs=raw_input,
            experiment=experiment,
        )
        all_preprocessed.append(preprocessed)

    return all_preprocessed


def general_pre_process_raw_inputs(
    raw_inputs: dict[str, Any],
    experiment: Union["Experiment", "ServeExperiment"],
) -> dict[str, torch.Tensor]:
    inputs_prepared_for_memory = {}
    for name, cur_input in raw_inputs.items():
        input_object = experiment.inputs[name]

        match input_object:
            case ComputedSequenceInputInfo():
                cur_input = input_object.encode_func(cur_input)

            case (
                ComputedTabularInputInfo()
                | ComputedPredictTabularInputInfo()
                | ComputedServeTabularInputInfo()
            ):
                cur_input = _impute_missing_tabular_values(
                    input_object=input_object,
                    inputs_values=cur_input,
                )

        inputs_prepared_for_memory[name] = cur_input

    inputs_prepared = prepare_inputs_memory(
        inputs=inputs_prepared_for_memory,
        inputs_objects=experiment.inputs,
        test_mode=True,
    )

    inputs_final = impute_missing_modalities_wrapper(
        inputs_values=inputs_prepared,
        inputs_objects=experiment.inputs,
    )

    return inputs_final


def _impute_missing_tabular_values(
    input_object: (
        ComputedTabularInputInfo
        | ComputedPredictTabularInputInfo
        | ComputedServeTabularInputInfo
    ),
    inputs_values: dict[str, Any],
) -> dict[str, Any]:
    # TODO: Implement
    return inputs_values



import math
from typing import Literal, Sequence, Tuple, Type

import numpy as np
from torch import nn

from eir.models.layers.lcl_layers import LCL, LCLResidualBlock


def get_1d_projection_layer(
    input_dimension: int,
    target_dimension: int,
    projection_layer_type: Literal[
        "auto",
        "lcl",
        "lcl_residual",
        "linear",
        "cnn",
    ] = "auto",
    lcl_diff_tolerance: int = 0,
) -> LCLResidualBlock | LCL | nn.Linear | nn.Identity:
    """
    Note: These currently (and possibly will only continue to) support
          projection from 1D to 1D.
    """
    if projection_layer_type == "auto":
        if input_dimension == target_dimension:
            return nn.Identity()

        lcl_residual_projection = get_lcl_projection_layer(
            input_dimension=input_dimension,
            target_dimension=target_dimension,
            layer_type="lcl_residual",
            diff_tolerance=lcl_diff_tolerance,
        )
        if lcl_residual_projection is not None:
            return lcl_residual_projection

        lcl_projection = get_lcl_projection_layer(
            input_dimension=input_dimension,
            target_dimension=target_dimension,
            layer_type="lcl",
            diff_tolerance=lcl_diff_tolerance,
        )
        if lcl_projection is not None:
            return lcl_projection

        return nn.Linear(
            in_features=input_dimension,
            out_features=target_dimension,
            bias=True,
        )

    elif projection_layer_type == "lcl_residual":
        layer = get_lcl_projection_layer(
            input_dimension=input_dimension,
            target_dimension=target_dimension,
            layer_type="lcl_residual",
            diff_tolerance=lcl_diff_tolerance,
        )
        if layer is None:
            raise ValueError(
                f"Cannot create lcl_residual projection layer for "
                f"input_dimension={input_dimension} and "
                f"target_dimension={target_dimension} for projection. "
                f"Try using projection_layer_type='auto'."
            )
        else:
            return layer

    elif projection_layer_type == "lcl" or projection_layer_type == "lcl_residual":
        layer = get_lcl_projection_layer(
            input_dimension=input_dimension,
            target_dimension=target_dimension,
            layer_type=projection_layer_type,
            diff_tolerance=lcl_diff_tolerance,
        )
        if layer is None:
            raise ValueError(
                f"Cannot create lcl projection layer for "
                f"input_dimension={input_dimension} and "
                f"target_dimension={target_dimension} for projection. "
                f"Try using projection_layer_type='auto'."
            )
        else:
            return layer

    elif projection_layer_type == "linear":
        if input_dimension == target_dimension:
            return nn.Identity()
        else:
            return nn.Linear(
                in_features=input_dimension,
                out_features=target_dimension,
                bias=True,
            )

    elif projection_layer_type == "cnn":
        raise NotImplementedError()

    else:
        raise ValueError(f"Invalid projection_layer_type: {projection_layer_type}")


def get_lcl_projection_layer(
    input_dimension: int,
    target_dimension: int,
    layer_type: Literal["lcl_residual", "lcl"] = "lcl_residual",
    kernel_width_candidates: Sequence[int] = tuple(range(1, 1024 + 1)),
    out_feature_sets_candidates: Sequence[int] = tuple(range(1, 512 + 1)),
    diff_tolerance: int = 0,
) -> LCLResidualBlock | LCL | None:
    layer_class: Type[LCLResidualBlock] | Type[LCL]
    match layer_type:
        case "lcl_residual":
            layer_class = LCLResidualBlock
            n_lcl_layers = 2
        case "lcl":
            layer_class = LCL
            n_lcl_layers = 1
        case _:
            raise ValueError(f"Unknown layer type: {layer_type}")

    search_func = _find_best_lcl_kernel_width_and_out_feature_sets
    solution = search_func(
        input_dimension=input_dimension,
        target_dimension=target_dimension,
        n_layers=n_lcl_layers,
        kernel_width_candidates=kernel_width_candidates,
        out_feature_sets_candidates=out_feature_sets_candidates,
        diff_tolerance=diff_tolerance,
    )

    if solution is None:
        return None

    best_kernel_size, best_out_feature_sets = solution
    best_layer = layer_class(
        in_features=input_dimension,
        kernel_size=best_kernel_size,
        out_feature_sets=best_out_feature_sets,
    )

    return best_layer


def _find_best_lcl_kernel_width_and_out_feature_sets(
    input_dimension: int,
    target_dimension: int,
    n_layers: int,
    kernel_width_candidates: Sequence[int] = tuple(range(1, 1024 + 1)),
    out_feature_sets_candidates: Sequence[int] = tuple(range(1, 64 + 1)),
    diff_tolerance: int = 0,
) -> Tuple[int, int] | None:
    best_diff = np.Inf
    best_kernel_width = None
    best_out_feature_sets = None

    def _compute(
        input_dimension_: int, kernel_width_: int, out_feature_sets_: int
    ) -> int:
        num_chunks_ = int(math.ceil(input_dimension_ / kernel_width_))
        out_features_ = num_chunks_ * out_feature_sets_
        return out_features_

    for out_feature_sets in out_feature_sets_candidates:
        for kernel_width in kernel_width_candidates:
            if kernel_width > input_dimension:
                continue

            out_features = input_dimension
            for n in range(n_layers):
                out_features = _compute(
                    input_dimension_=out_features,
                    kernel_width_=kernel_width,
                    out_feature_sets_=out_feature_sets,
                )

            if out_features < target_dimension:
                continue

            diff = abs(out_features - target_dimension)

            if diff < best_diff:
                best_diff = diff
                best_kernel_width = kernel_width
                best_out_feature_sets = out_feature_sets

            if diff <= diff_tolerance:
                break

    if best_diff != 0:
        return None

    if best_kernel_width is None:
        return None

    assert best_kernel_width is not None
    assert best_out_feature_sets is not None

    return best_kernel_width, best_out_feature_sets



from typing import Literal, Tuple

import torch
from einops import rearrange
from torch import nn
from torchvision.ops import StochasticDepth

from eir.models.layers.attention_layers import LinearAttention
from eir.models.layers.norm_layers import GRN
from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


class SEBlock(nn.Module):
    def __init__(self, channels: int, reduction: int):
        super(SEBlock, self).__init__()
        reduced_channels = max(1, channels // reduction)
        self.avg_pool = nn.AdaptiveAvgPool2d(1)

        self.conv_down = nn.Conv2d(
            in_channels=channels,
            out_channels=reduced_channels,
            kernel_size=1,
            padding=0,
            bias=True,
        )
        self.act_1 = nn.GELU()

        self.conv_up = nn.Conv2d(
            in_channels=reduced_channels,
            out_channels=channels,
            kernel_size=1,
            padding=0,
            bias=True,
        )

        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        out = self.avg_pool(x)

        out = self.conv_down(out)
        out = self.act_1(out)

        out = self.conv_up(out)
        out = self.sigmoid(out)

        return out


class ConvAttentionBlock(nn.Module):
    def __init__(
        self,
        channels: int,
        height: int,
        width: int,
        num_heads: int = 4,
        dropout_p: float = 0.1,
        attention_mode: Literal["spatial", "channel"] = "spatial",
        attention_type: Literal["full", "linear"] = "full",
    ):
        super().__init__()
        self.in_channels = channels
        self.out_channels = channels
        self.in_height = height
        self.in_width = width
        self.attention_mode = attention_mode
        self.attention_type = attention_type

        self.embedding_dim = channels if attention_mode == "spatial" else height * width
        self.num_heads = adjust_num_heads(
            num_heads=num_heads,
            embedding_dim=self.embedding_dim,
        )

        self.attention: nn.MultiheadAttention | LinearAttention
        if attention_type == "full":
            self.norm = nn.LayerNorm(normalized_shape=self.embedding_dim)
            self.attention = nn.MultiheadAttention(
                embed_dim=self.embedding_dim,
                num_heads=self.num_heads,
                batch_first=True,
                dropout=dropout_p,
            )
        elif attention_type == "linear":
            self.norm = nn.LayerNorm(normalized_shape=[channels, height, width])
            self.attention = LinearAttention(
                embed_dim=self.embedding_dim,
                heads=self.num_heads,
                dim_head=self.embedding_dim // self.num_heads,
            )

        self.grn = GRN(in_channels=channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        batch_size, channels, height, width = x.size()

        if self.attention_type == "full":
            if self.attention_mode == "spatial":
                out = rearrange(x, "b c h w -> b (h w) c")
            elif self.attention_mode == "channel":
                out = rearrange(x, "b c h w -> b (h w) c").permute(0, 2, 1)
            else:
                raise ValueError()

            out = self.norm(out)
            attn_output, _ = self.attention(out, out, out)
            out = out + attn_output

            if self.attention_mode == "spatial":
                out = rearrange(out, "b (h w) c -> b c h w", h=height, w=width)
            elif self.attention_mode == "channel":
                out = rearrange(out, "b c (h w) -> b c h w", h=height, w=width)
            else:
                raise ValueError()

        elif self.attention_type == "linear":
            assert self.attention_mode == "spatial"
            out = self.norm(x)
            attn_output = self.attention(x)
            out = out + attn_output

        else:
            raise ValueError()

        out = self.grn(out)

        return x + out


def adjust_num_heads(num_heads: int, embedding_dim: int) -> int:
    while embedding_dim % num_heads != -0:
        num_heads -= 1

    logger.debug(
        f"Adjusted base number of heads to {num_heads} "
        f"according to embedding dim {embedding_dim}."
    )
    return num_heads


class CNNResidualBlockBase(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        rb_do: float,
        dilation_w: int,
        dilation_h: int,
        conv_1_kernel_h: int = 1,
        conv_1_kernel_w: int = 12,
        conv_1_padding_w: int = 4,
        conv_1_padding_h: int = 4,
        down_stride_w: int = 4,
        down_stride_h: int = 1,
        stochastic_depth_p: float = 0.0,
        conv_downsample_identity: bool = True,
    ):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.conv_1_kernel_h = conv_1_kernel_h
        self.conv_1_kernel_w = conv_1_kernel_w

        self.dilation_w = dilation_w
        self.dilation_h = dilation_h

        self.conv_1_padding_w = conv_1_padding_w
        self.conv_1_padding_h = conv_1_padding_h

        self.down_stride_w = down_stride_w
        self.down_stride_h = down_stride_h

        self.stochastic_depth_p = stochastic_depth_p

        self.conv_downsample_identity = conv_downsample_identity

        self.rb_do = nn.Dropout2d(rb_do)

        self.conv_ds = nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=self.in_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
            groups=in_channels,
        )

        self.act_1 = nn.GELU()

        self.norm_1 = nn.GroupNorm(num_groups=1, num_channels=in_channels)
        self.conv_1 = nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=(self.conv_1_kernel_h, self.conv_1_kernel_w),
            stride=(self.down_stride_h, self.down_stride_w),
            padding=(self.conv_1_padding_h, self.conv_1_padding_w),
            dilation=(self.dilation_h, self.dilation_w),
            bias=True,
        )

        conv_2_kernel_h, conv_2_padding_h = _compute_conv_2_parameters(
            conv_1_kernel_size=conv_1_kernel_h, dilation=dilation_h
        )

        conv_2_kernel_w, conv_2_padding_w = _compute_conv_2_parameters(
            conv_1_kernel_size=conv_1_kernel_w, dilation=dilation_w
        )

        self.grn = GRN(in_channels=out_channels)

        self.conv_2 = nn.Conv2d(
            in_channels=self.out_channels,
            out_channels=self.out_channels,
            kernel_size=(conv_2_kernel_h, conv_2_kernel_w),
            stride=(1, 1),
            padding=(conv_2_padding_h, conv_2_padding_w),
            dilation=(dilation_h, dilation_w),
            bias=True,
        )

        self.downsample_identity: nn.Module = nn.Identity()
        if self.conv_downsample_identity:
            self.downsample_identity = nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=(self.conv_1_kernel_h, self.conv_1_kernel_w),
                stride=(self.down_stride_h, self.down_stride_w),
                padding=(self.conv_1_padding_h, self.conv_1_padding_w),
                dilation=(self.dilation_h, self.dilation_w),
                bias=True,
            )

        self.stochastic_depth = StochasticDepth(p=self.stochastic_depth_p, mode="batch")

        self.se_block = SEBlock(channels=out_channels, reduction=16)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()


def _compute_conv_2_parameters(
    conv_1_kernel_size: int,
    dilation: int,
) -> Tuple[int, int]:
    conv_2_kernel = (
        conv_1_kernel_size - 1 if conv_1_kernel_size % 2 == 0 else conv_1_kernel_size
    )

    conv_2_padding = conv_2_kernel // 2
    conv_2_padding = conv_2_padding * dilation

    return conv_2_kernel, conv_2_padding


class FirstCNNBlock(CNNResidualBlockBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        delattr(self, "conv_ds")
        delattr(self, "norm_1")
        delattr(self, "downsample_identity")
        delattr(self, "act_1")
        delattr(self, "grn")
        delattr(self, "rb_do")
        delattr(self, "conv_2")
        delattr(self, "se_block")
        delattr(self, "stochastic_depth")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.conv_1(x)

        return out


class CNNResidualBlock(CNNResidualBlockBase):
    def __init__(self, full_preact: bool = False, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.full_preact = full_preact

    def forward(self, x: torch.Tensor) -> torch.Tensor:

        out = self.conv_ds(x)

        out = self.norm_1(out)

        if self.full_preact:
            identity = self.downsample_identity(out)
        else:
            identity = self.downsample_identity(x)

        out = self.conv_1(out)

        out = self.act_1(out)
        out = self.grn(out)
        out = self.rb_do(out)
        out = self.conv_2(out)

        channel_recalibrations = self.se_block(out)
        out = out * channel_recalibrations

        out = self.stochastic_depth(out)

        out = out + identity

        return out


class DownSamplingResidualBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        in_height: int,
        in_width: int,
    ):
        """
        Note: Slightly different approach here compared to the upsampling below,
        here we do the downsampling in one-go with a strided convolution.
        """
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = in_channels
        self.in_height = in_height
        self.in_width = in_width

        self.stride_h, self.stride_w = _compute_params_for_down_sampling(
            cur_height=in_height,
            cur_width=in_width,
        )

        self.norm_1 = nn.GroupNorm(num_groups=1, num_channels=in_channels)
        self.act_1 = nn.GELU()

        self.conv_1 = nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=3,
            stride=(self.stride_h, self.stride_w),
            padding=1,
            bias=True,
        )

        self.identity = nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=3,
            stride=(self.stride_h, self.stride_w),
            padding=1,
            bias=True,
        )

        self.grn = GRN(in_channels=self.out_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = self.identity(x)

        out = self.norm_1(x)
        out = self.act_1(out)
        out = self.conv_1(out)
        out = self.grn(out)

        return out + identity


def _compute_params_for_down_sampling(
    cur_height: int, cur_width: int
) -> tuple[int, int]:

    if cur_height == 1:
        stride_h = 1
    else:
        stride_h = 2

    if cur_width == 1:
        stride_w = 1
    else:
        stride_w = 2

    return stride_h, stride_w


class UpSamplingResidualBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        in_height: int,
        in_width: int,
        upsample_height: bool = True,
        upsample_width: bool = True,
    ):
        super(UpSamplingResidualBlock, self).__init__()
        """
        Note: Always applying a Conv to the upsampled identity seems to help
        stabilize training.
        """

        self.in_channels = in_channels
        self.out_channels = in_channels
        self.in_height = in_height
        self.in_width = in_width
        self.upsample_height = upsample_height
        self.upsample_width = upsample_width

        self.stride_h, self.stride_w = _compute_params_for_up_sampling(
            upsample_height=upsample_height,
            upsample_width=upsample_width,
        )

        self.upsample = nn.Upsample(
            scale_factor=(self.stride_h, self.stride_w),
            mode="bilinear",
            align_corners=True,
        )

        self.norm_1 = nn.GroupNorm(num_groups=1, num_channels=in_channels)
        self.act_1 = nn.GELU()

        self.conv_1 = nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
        )

        self.identity = nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=True,
        )

        self.grn = GRN(in_channels=self.out_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        upsampled = self.upsample(x)
        identity = self.identity(upsampled)

        out = upsampled
        out = self.norm_1(out)
        out = self.act_1(out)
        out = self.conv_1(out)
        out = self.grn(out)

        return out + identity


def _compute_params_for_up_sampling(
    upsample_height: bool,
    upsample_width: bool,
):
    stride_h = 2 if upsample_height else 1
    stride_w = 2 if upsample_width else 1
    return stride_h, stride_w



from typing import Dict, Tuple

import torch
from torch._C._nn import pad

from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.train_utils.evaluation_handlers.evaluation_handlers_utils import (
    SpecialTokens,
    get_special_tokens,
)


def prepare_sequence_input_for_sequence_output(
    input_object: ComputedSequenceInputInfo,
    cur_seq: torch.Tensor,
    input_name: str,
    device: str,
) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
    assert input_object.tokenizer is not None
    special_tokens = get_special_tokens(
        tokenizer=input_object.tokenizer,
        vocab=input_object.vocab,
    )

    cur_seq, cur_target = sample_autoregressive_batch(
        batch_tensor=cur_seq,
        batch_size=cur_seq.shape[0],
        special_tokens=special_tokens,
    )
    assert cur_seq.shape == cur_target.shape
    assert cur_seq.dim() == 2
    assert cur_seq.shape[1] == input_object.computed_max_length

    cur_seq = cur_seq.to(device=device)

    cur_target_dict = {input_name: cur_target.to(device=device)}

    return cur_seq, cur_target_dict


def sample_autoregressive_batch(
    batch_tensor: torch.Tensor,
    batch_size: int,
    special_tokens: SpecialTokens,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    The reason for padding with the BOS token is that the tensor we get here
    is already at max_length. If we had e.g. a full, long sequence, we could
    simply slice that directly (+1 for the target), but here we need to pad
    the input at the beginning, opting for a BOS token.
    """
    st = special_tokens

    inputs = []
    targets = []

    batch_tensor = pad_batch_with_bos(
        batch_tensor=batch_tensor,
        bos_value=st.bos_idx,
    )

    for idx in range(batch_size):
        cur_sample = batch_tensor[idx]

        cur_sample = _switch_first_pad_with_eos(
            sample=cur_sample,
            pad_value=st.pad_idx,
            eos_value=st.eos_idx,
        )

        cur_target = cur_sample[1:]
        cur_sample = cur_sample[:-1]

        inputs.append(cur_sample)
        targets.append(cur_target)

    inputs_tensor = torch.stack(tensors=inputs)
    target_tensor = torch.stack(tensors=targets)
    return inputs_tensor, target_tensor


def _switch_first_pad_with_eos(
    sample: torch.Tensor,
    pad_value: int,
    eos_value: int,
) -> torch.Tensor:
    first_pad_match = (sample == pad_value).nonzero(as_tuple=False)
    if len(first_pad_match) != 0:
        first_pad_index = first_pad_match[0]
        sample[first_pad_index] = eos_value

    return sample


def pad_batch_with_bos(
    batch_tensor: torch.Tensor,
    bos_value: int,
) -> torch.Tensor:
    left_padding = 1
    batch_tensor = pad(input=batch_tensor, pad=[left_padding, 0], value=bos_value)

    return batch_tensor



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Sequence

import numpy as np
import pandas as pd
import torch
from sklearn.preprocessing import StandardScaler

from eir.data_load import label_setup
from eir.data_load.data_source_modules.deeplake_ops import is_deeplake_dataset
from eir.data_load.label_setup import (
    TabularFileInfo,
    al_all_column_ops,
    al_label_dict,
    al_label_transformers,
    al_target_label_dict,
    transform_label_df,
)
from eir.experiment_io.experiment_io import load_transformers
from eir.predict_modules.predict_tabular_input_setup import prep_missing_con_dict
from eir.setup.config import Configs
from eir.setup.schemas import (
    ArrayOutputTypeConfig,
    ImageOutputTypeConfig,
    OutputConfig,
    SequenceOutputTypeConfig,
    TabularOutputTypeConfig,
)
from eir.target_setup.target_label_setup import (
    MissingTargetsInfo,
    compute_missing_ids_per_tabular_output,
    convert_dtypes,
    df_to_nested_dict,
    gather_data_pointers_from_data_source,
    gather_torch_nan_missing_ids,
    get_all_output_and_target_names,
    get_missing_targets_info,
    get_tabular_target_file_infos,
)


@dataclass
class PredictTargetLabels:
    predict_labels: pd.DataFrame
    label_transformers: Dict[str, al_label_transformers]

    @property
    def all_labels(self):
        return self.predict_labels


@dataclass
class MergedPredictTargetLabels:
    label_dict: al_target_label_dict
    label_transformers: dict[str, dict[str, al_label_transformers]]
    missing_ids_per_output: MissingTargetsInfo

    @property
    def all_labels(self):
        return self.label_dict


def get_tabular_target_labels_for_predict(
    output_folder: str,
    tabular_info: TabularFileInfo,
    output_name: str,
    custom_column_label_parsing_ops: al_all_column_ops,
    ids: Sequence[str],
) -> PredictTargetLabels:
    all_columns = list(tabular_info.cat_columns) + list(tabular_info.con_columns)
    if not all_columns:
        raise ValueError(f"No columns specified in {tabular_info}.")

    con_columns = list(tabular_info.con_columns)
    cat_columns = list(tabular_info.cat_columns)

    df_labels_test = _load_labels_for_predict(
        tabular_info=tabular_info,
        ids_to_keep=ids,
        custom_label_ops=custom_column_label_parsing_ops,
    )

    label_transformers = load_transformers(
        output_folder=output_folder,
        transformers_to_load={output_name: all_columns},
    )

    labels_dict = parse_labels_for_predict(
        con_columns=con_columns,
        cat_columns=cat_columns,
        df_labels_test=df_labels_test,
        label_transformers=label_transformers[output_name],
    )

    labels = PredictTargetLabels(
        predict_labels=labels_dict,
        label_transformers=label_transformers,
    )

    return labels


def _load_labels_for_predict(
    tabular_info: TabularFileInfo,
    ids_to_keep: Sequence[str],
    custom_label_ops: al_all_column_ops = None,
) -> pd.DataFrame:
    parse_wrapper = label_setup.get_label_parsing_wrapper(
        label_parsing_chunk_size=tabular_info.parsing_chunk_size
    )
    df_labels_test = parse_wrapper(
        label_file_tabular_info=tabular_info,
        ids_to_keep=ids_to_keep,
        custom_label_ops=custom_label_ops,
    )

    return df_labels_test


def parse_labels_for_predict(
    con_columns: Sequence[str],
    cat_columns: Sequence[str],
    df_labels_test: pd.DataFrame,
    label_transformers: al_label_transformers,
) -> pd.DataFrame:
    con_transformers = _extract_target_con_transformers(
        label_transformers=label_transformers,
        con_columns=con_columns,
    )
    train_con_column_means = prep_missing_con_dict(con_transformers=con_transformers)

    df_labels_test = label_setup.handle_missing_label_values_in_df(
        df=df_labels_test,
        cat_label_columns=cat_columns,
        con_label_columns=con_columns,
        con_manual_values=train_con_column_means,
        name="test_df",
        impute_missing=False,
    )

    assert len(label_transformers) > 0
    df_labels_test = transform_label_df(
        df_labels=df_labels_test,
        label_transformers=label_transformers,
        impute_missing=False,
    )

    return df_labels_test


def _extract_target_con_transformers(
    label_transformers: al_label_transformers,
    con_columns: Sequence[str],
) -> Dict[str, StandardScaler]:
    con_transformers = {}

    for target_column, transformer_object in label_transformers.items():
        if target_column not in con_columns:
            continue

        assert target_column not in con_transformers
        assert isinstance(transformer_object, StandardScaler)

        con_transformers[target_column] = transformer_object

    assert len(con_transformers) == len(con_columns)

    return con_transformers


def get_target_labels_for_testing(
    configs_overloaded_for_predict: Configs,
    custom_column_label_parsing_ops: al_all_column_ops,
    ids: Sequence[str],
) -> MergedPredictTargetLabels:
    pc = configs_overloaded_for_predict
    output_configs = pc.output_configs

    df_labels_test = pd.DataFrame(index=list(ids))
    tabular_label_transformers = {}

    all_ids: set[str] = set(ids)
    per_modality_missing_ids: dict[str, set[str]] = {}
    within_modality_missing_ids: dict[str, dict[str, set[str]]] = {}

    dtypes: dict[str, dict[str, Any]] = {}

    tabular_target_files_info = get_tabular_target_file_infos(
        output_configs=pc.output_configs
    )

    for output_config in output_configs:
        output_source = output_config.output_info.output_source
        output_type_info = output_config.output_type_info
        output_name = output_config.output_info.output_name

        match output_type_info:
            case TabularOutputTypeConfig():
                cur_tabular_info = tabular_target_files_info[output_name]
                cur_labels = get_tabular_target_labels_for_predict(
                    output_folder=pc.global_config.basic_experiment.output_folder,
                    tabular_info=cur_tabular_info,
                    output_name=output_name,
                    custom_column_label_parsing_ops=custom_column_label_parsing_ops,
                    ids=ids,
                )
                tabular_label_transformers[output_name] = cur_labels.label_transformers

                all_labels = cur_labels.all_labels
                cur_ids = set(all_labels.index)
                missing_ids = all_ids.difference(cur_ids)
                per_modality_missing_ids[output_name] = missing_ids

                missing_ids_per_target_column = compute_missing_ids_per_tabular_output(
                    all_labels_df=all_labels,
                    tabular_info=cur_tabular_info,
                    output_name=output_name,
                )
                within_modality_missing_ids = {
                    **within_modality_missing_ids,
                    **missing_ids_per_target_column,
                }

            case SequenceOutputTypeConfig():
                cur_labels = set_up_delayed_predict_target_labels(
                    ids=ids,
                    output_name=output_name,
                )

                cur_missing_ids = gather_torch_nan_missing_ids(
                    labels=cur_labels.all_labels,
                    output_name=output_name,
                )

                per_modality_missing_ids[output_name] = cur_missing_ids

            case ArrayOutputTypeConfig() | ImageOutputTypeConfig():
                cur_labels = _set_up_predict_file_target_labels(
                    ids=ids,
                    output_config=output_config,
                )

                cur_missing_ids = gather_torch_nan_missing_ids(
                    labels=cur_labels.all_labels,
                    output_name=output_name,
                )
                per_modality_missing_ids[output_name] = cur_missing_ids

                # this is needed as having missing modalities in deeplake
                # will cause conversion of int64 deeplake pointers to float64
                is_deeplake = is_deeplake_dataset(data_source=output_source)
                if is_deeplake:
                    dtypes[output_name] = {output_name: np.dtype("int64")}
                else:
                    dtypes[output_name] = {output_name: np.dtype("O")}

            case _:
                raise NotImplementedError(
                    f"Output type {output_type_info} not implemented"
                )

        df_labels_cur = cur_labels.predict_labels

        df_labels_cur["Output Name"] = output_name

        if output_name not in dtypes:
            dtypes[output_name] = df_labels_cur.dtypes.to_dict()

        df_labels_test = pd.concat((df_labels_test, df_labels_cur))

    df_labels_test = df_labels_test.set_index("Output Name", append=True)

    df_labels_test = df_labels_test.dropna(how="all")

    dtypes = convert_dtypes(dtypes=dtypes)
    test_labels_dict = df_to_nested_dict(df=df_labels_test, dtypes=dtypes)

    outputs_and_targets = get_all_output_and_target_names(output_configs=output_configs)
    missing_target_info = get_missing_targets_info(
        missing_ids_per_modality=per_modality_missing_ids,
        missing_ids_within_modality=within_modality_missing_ids,
        output_and_target_names=outputs_and_targets,
    )

    test_labels_object = MergedPredictTargetLabels(
        label_dict=test_labels_dict,
        label_transformers=tabular_label_transformers,
        missing_ids_per_output=missing_target_info,
    )

    return test_labels_object


def _set_up_predict_file_target_labels(
    ids: Sequence[str],
    output_config: OutputConfig,
) -> PredictTargetLabels:
    output_name = output_config.output_info.output_name
    output_source = output_config.output_info.output_source
    output_inner_key = output_config.output_info.output_inner_key

    id_to_data_pointer_mapping = gather_data_pointers_from_data_source(
        data_source=Path(output_source),
        validate=True,
        output_inner_key=output_inner_key,
    )

    ids_set = set(ids)
    labels: al_label_dict = {
        id_: {output_name: id_to_data_pointer_mapping[id_]} for id_ in ids_set
    }

    df_labels = pd.DataFrame.from_dict(labels, orient="index")
    df_labels["Output Name"] = output_name

    return PredictTargetLabels(
        predict_labels=df_labels,
        label_transformers={},
    )


def set_up_delayed_predict_target_labels(
    ids: Sequence[str],
    output_name: str,
) -> PredictTargetLabels:
    ids_set = set(ids)
    labels: al_label_dict = {id_: {output_name: torch.nan} for id_ in ids_set}

    df_labels = pd.DataFrame.from_dict(labels, orient="index")
    df_labels["Output Name"] = output_name

    return PredictTargetLabels(
        predict_labels=df_labels,
        label_transformers={},
    )



from dataclasses import dataclass
from typing import TYPE_CHECKING

import torch
from torch import nn

if TYPE_CHECKING:
    from eir.setup.input_setup_modules.common import DataDimensions


@dataclass
class LinearModelConfig:
    """
    :param fc_repr_dim:
        Number of output nodes in the first and only hidden layer.
    :param l1:
        L1 regularisation to apply to the first layer.
    """

    fc_repr_dim: int = 32
    l1: float = 0.0


class LinearModel(nn.Module):
    def __init__(
        self, model_config: LinearModelConfig, data_dimensions: "DataDimensions"
    ):
        super().__init__()

        self.model_config = model_config
        self.data_dimensions = data_dimensions

        self.fc_0 = nn.Linear(
            self.fc_1_in_features,
            self.model_config.fc_repr_dim,
            bias=True,
        )

        self._init_weights()

    @property
    def fc_1_in_features(self) -> int:
        return self.data_dimensions.num_elements()

    @property
    def l1_penalized_weights(self) -> torch.Tensor:
        return self.fc_0.weight

    @property
    def num_out_features(self) -> int:
        return self.model_config.fc_repr_dim

    def _init_weights(self):
        pass

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        out = input.view(input.shape[0], -1)

        out = self.fc_0(out)

        return out



from typing import Union

from eir.data_load import datasets
from eir.data_load.datasets import al_local_datasets
from eir.data_load.label_setup import al_target_label_dict
from eir.setup.config import Configs
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.output_setup import al_output_objects_as_dict
from eir.target_setup.target_label_setup import MissingTargetsInfo


def set_up_default_dataset(
    configs: Configs,
    target_labels_dict: Union[None, al_target_label_dict],
    inputs_as_dict: al_input_objects_as_dict,
    outputs_as_dict: al_output_objects_as_dict,
    missing_ids_per_output: MissingTargetsInfo,
) -> al_local_datasets:
    test_dataset_kwargs = datasets.construct_default_dataset_kwargs_from_cl_args(
        target_labels_dict=target_labels_dict,
        outputs=outputs_as_dict,
        inputs=inputs_as_dict,
        test_mode=True,
        missing_ids_per_output=missing_ids_per_output,
    )

    test_dataset = datasets.DiskDataset(**test_dataset_kwargs)

    return test_dataset



from eir.models.output.array.array_output_modules import ArrayOutputWrapperModule
from eir.models.output.sequence.sequence_output_modules import SequenceOutputModule
from eir.models.output.tabular.linear import LinearOutputModule
from eir.models.output.tabular.mlp_residual import ResidualMLPOutputModule

al_output_modules = (
    ResidualMLPOutputModule
    | LinearOutputModule
    | SequenceOutputModule
    | ArrayOutputWrapperModule
)



from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, Sequence

import pandas as pd

from eir.data_load import label_setup
from eir.data_load.label_setup import (
    TabularFileInfo,
    al_all_column_ops,
    al_label_transformers,
    transform_label_df,
)
from eir.experiment_io.experiment_io import load_transformers
from eir.setup import schemas
from eir.setup.input_setup_modules import setup_tabular

if TYPE_CHECKING:
    from eir.train_utils.step_logic import Hooks


@dataclass
class ComputedPredictTabularInputInfo:
    labels: "PredictInputLabels"
    input_config: schemas.InputConfig


@dataclass
class PredictInputLabels:
    predict_labels: pd.DataFrame
    label_transformers: al_label_transformers

    @property
    def all_labels(self):
        return self.predict_labels


def setup_tabular_input_for_testing(
    input_config: schemas.InputConfig,
    ids: Sequence[str],
    hooks: "Hooks",
    output_folder: str,
) -> ComputedPredictTabularInputInfo:
    input_type_info = input_config.input_type_info
    assert isinstance(input_type_info, schemas.TabularInputDataConfig)

    tabular_file_info = setup_tabular.get_tabular_input_file_info(
        input_source=input_config.input_info.input_source,
        tabular_data_type_config=input_type_info,
    )

    custom_ops = hooks.custom_column_label_parsing_ops if hooks else None
    predict_labels = get_input_labels_for_predict(
        tabular_file_info=tabular_file_info,
        input_name=input_config.input_info.input_name,
        custom_label_ops=custom_ops,
        ids=ids,
        output_folder=output_folder,
    )

    predict_input_info = ComputedPredictTabularInputInfo(
        labels=predict_labels, input_config=input_config
    )

    return predict_input_info


def get_input_labels_for_predict(
    tabular_file_info: TabularFileInfo,
    input_name: str,
    custom_label_ops: al_all_column_ops,
    ids: Sequence[str],
    output_folder: str,
) -> PredictInputLabels:
    if len(tabular_file_info.con_columns) + len(tabular_file_info.cat_columns) < 1:
        raise ValueError(f"No label columns specified in {tabular_file_info}.")

    parse_wrapper = label_setup.get_label_parsing_wrapper(
        label_parsing_chunk_size=tabular_file_info.parsing_chunk_size
    )
    df_labels_test = parse_wrapper(
        label_file_tabular_info=tabular_file_info,
        ids_to_keep=ids,
        custom_label_ops=custom_label_ops,
    )

    label_setup.pre_check_label_df(df=df_labels_test, name="Testing DataFrame")

    all_columns = list(tabular_file_info.con_columns) + list(
        tabular_file_info.cat_columns
    )
    label_transformers_with_input_name = load_transformers(
        transformers_to_load={input_name: all_columns}, output_folder=output_folder
    )
    loaded_fit_label_transformers = label_transformers_with_input_name[input_name]

    con_transformers = _extract_input_con_transformers(
        loaded_fit_label_transformers=loaded_fit_label_transformers,
        con_columns=tabular_file_info.con_columns,
    )
    train_con_column_means = prep_missing_con_dict(con_transformers=con_transformers)

    df_labels_test_no_na = label_setup.handle_missing_label_values_in_df(
        df=df_labels_test,
        cat_label_columns=tabular_file_info.cat_columns,
        con_label_columns=tabular_file_info.con_columns,
        con_manual_values=train_con_column_means,
        name="test_df",
        impute_missing=True,
    )

    df_labels_test_final = transform_label_df(
        df_labels=df_labels_test_no_na,
        label_transformers=loaded_fit_label_transformers,
        impute_missing=True,
    )

    labels_data_object = PredictInputLabels(
        predict_labels=df_labels_test_final,
        label_transformers=loaded_fit_label_transformers,
    )

    return labels_data_object


def _extract_input_con_transformers(
    loaded_fit_label_transformers, con_columns: Sequence[str]
):
    con_transformers = {
        k: v for k, v in loaded_fit_label_transformers.items() if k in con_columns
    }

    assert len(con_transformers) == len(con_columns)

    return con_transformers


def prep_missing_con_dict(con_transformers: al_label_transformers) -> Dict[str, float]:
    train_means = {
        column: transformer.mean_[0] for column, transformer in con_transformers.items()
    }

    return train_means



import base64
from typing import TYPE_CHECKING, Any

import numpy as np
import torch
from scipy.special import softmax
from sklearn.preprocessing import StandardScaler

from eir.models.model_setup_modules.input_model_setup.input_model_setup_sequence import (  # noqa
    get_sequence_model,
)
from eir.models.model_setup_modules.output_model_setup_modules.output_model_setup_array import (  # noqa
    get_array_or_image_output_module_from_model_config,
)
from eir.models.model_setup_modules.output_model_setup_modules.output_model_setup_sequence import (  # noqa
    get_sequence_output_module_from_model_config,
)
from eir.models.model_setup_modules.output_model_setup_modules.output_model_setup_tabular import (  # noqa
    get_tabular_output_module_from_model_config,
)
from eir.models.model_training_utils import recursive_to_device
from eir.models.output.sequence.sequence_output_modules import (
    SequenceOutputModuleConfig,
)
from eir.setup import schemas
from eir.setup.config_setup_modules.config_setup_utils import object_to_primitives
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.setup.output_setup import al_output_objects_as_dict
from eir.setup.output_setup_modules.array_output_setup import ComputedArrayOutputInfo
from eir.setup.output_setup_modules.image_output_setup import ComputedImageOutputInfo
from eir.setup.output_setup_modules.sequence_output_setup import (
    ComputedSequenceOutputInfo,
)
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)
from eir.train_utils.evaluation_handlers.evaluation_handlers_utils import (
    remove_special_tokens,
)
from eir.train_utils.evaluation_handlers.train_handlers_sequence_output import (
    decode_tokens,
    get_special_tokens,
    remove_special_tokens_from_string,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    pass

logger = get_logger(name=__name__)


def general_post_process(
    outputs: dict[str, dict[str, torch.Tensor | list[int] | np.ndarray]],
    output_objects: al_output_objects_as_dict,
    input_objects: al_input_objects_as_dict,
) -> dict[str, Any]:
    """
    Note that we always expect two levels for the predictions, in the case
    of tabular that's the output name and the column name, in the case of
    sequence we simply output name twice to conform to the same structure.

    Generally:
        - tabular outputs: torch.Tensor
        - sequence outputs: list[int]
        - array outputs: np.ndarray
    """
    post_processed: dict[str, Any] = {}

    outputs = recursive_to_device(obj=outputs, device="cpu")

    for output_name, output_object in output_objects.items():
        output_model_config = output_object.output_config.model_config
        cur_model_outputs = outputs[output_name]

        match output_object:
            case ComputedTabularOutputInfo():
                assert isinstance(
                    output_model_config, schemas.TabularOutputModuleConfig
                )

                tabular_outputs = _ensure_streamlined_tabular_values(
                    tabular_model_outputs=cur_model_outputs
                )

                tabular_inverse_transformed = _post_process_tabular_output(
                    output_object=output_object,
                    tabular_outputs=tabular_outputs,
                )
                post_processed[output_name] = tabular_inverse_transformed

            case ComputedSequenceOutputInfo():
                assert isinstance(output_model_config, SequenceOutputModuleConfig)

                output_type_info = output_object.output_config.output_type_info
                assert isinstance(output_type_info, schemas.SequenceOutputTypeConfig)

                generated_tokens = cur_model_outputs[output_name]
                cur_input_object = input_objects[output_name]
                assert isinstance(cur_input_object, ComputedSequenceInputInfo)
                assert cur_input_object.tokenizer is not None

                special_tokens = get_special_tokens(
                    tokenizer=cur_input_object.tokenizer,
                    vocab=cur_input_object.vocab,
                )

                assert isinstance(generated_tokens, list)
                generated_tokens = remove_special_tokens(
                    tokens=generated_tokens,
                    special_tokens=special_tokens,
                )

                generated_sample = decode_tokens(
                    tokens=generated_tokens,
                    vocab=cur_input_object.vocab,
                    split_on=output_type_info.split_on,
                )

                generated_sample = remove_special_tokens_from_string(
                    string=generated_sample,
                    special_tokens=special_tokens,
                )

                post_processed[output_name] = generated_sample

            case ComputedArrayOutputInfo():
                assert isinstance(output_model_config, schemas.ArrayOutputModuleConfig)

                array_np = cur_model_outputs[output_name]
                assert isinstance(array_np, np.ndarray)

                array_base64 = _post_process_array_outputs(array=array_np)
                post_processed[output_name] = array_base64

            case ComputedImageOutputInfo():
                assert isinstance(output_model_config, schemas.ArrayOutputModuleConfig)

                array_np = cur_model_outputs[output_name]
                assert isinstance(array_np, np.ndarray)

                array_base64 = _post_process_array_outputs(array=array_np)
                post_processed[output_name] = array_base64

            case _:
                raise NotImplementedError(
                    "Only tabular, sequence and array outputs are supported, got %s",
                    output_object,
                )

    post_processed = object_to_primitives(obj=post_processed)

    return post_processed


def _ensure_streamlined_tabular_values(
    tabular_model_outputs: dict[str, torch.Tensor | list[int] | np.ndarray]
) -> dict[str, torch.Tensor]:
    tensor_outputs = {
        k: v for k, v in tabular_model_outputs.items() if isinstance(v, torch.Tensor)
    }

    assert len(tensor_outputs) == len(tabular_model_outputs)

    return tensor_outputs


def _post_process_tabular_output(
    output_object: ComputedTabularOutputInfo, tabular_outputs: dict[str, torch.Tensor]
) -> dict[str, dict[str, float]]:
    processed_outputs: dict[str, dict[str, float]] = {}
    target_columns = output_object.target_columns

    for target_column_type, list_of_cols_of_this_type in target_columns.items():
        for cur_column in list_of_cols_of_this_type:
            if target_column_type == "con":
                cur_classes = [cur_column]
                cur_transformer = output_object.target_transformers[cur_column]
                cur_output_normalized = _normalize_continuous_outputs(
                    outputs=tabular_outputs[cur_column], transformer=cur_transformer
                )
                cur_output_normalized = cur_output_normalized
            elif target_column_type == "cat":
                cur_classes = output_object.target_transformers[cur_column].classes_
                cur_output_normalized = _normalize_categorical_outputs(
                    outputs=tabular_outputs[cur_column]
                )
            else:
                raise ValueError(
                    "Expected column type to be con or cat, but got %s",
                    target_column_type,
                )

            cur_class_to_confidence_mapping = {
                k: v for k, v in zip(cur_classes, cur_output_normalized)
            }

            processed_outputs[cur_column] = cur_class_to_confidence_mapping

    return processed_outputs


def _normalize_categorical_outputs(outputs: torch.Tensor) -> tuple[float]:
    cur_outputs_normalized = softmax(outputs).squeeze().tolist()

    return tuple(cur_outputs_normalized)


def _normalize_continuous_outputs(
    outputs: torch.Tensor, transformer: StandardScaler
) -> tuple[float]:
    cur_output_reshaped = outputs.reshape(1, -1)
    transform_func = transformer.inverse_transform
    cur_output_normalized = transform_func(cur_output_reshaped).squeeze()
    cur_outputs_normalized = (cur_output_normalized.item(),)

    return cur_outputs_normalized


def _post_process_array_outputs(array: np.ndarray) -> str:
    base64_encoded = _serialize_array_to_base64(array=array)
    return base64_encoded


def _serialize_array_to_base64(array: np.ndarray) -> str:
    array_bytes = array.tobytes()
    base64_encoded = base64.b64encode(array_bytes).decode("utf-8")
    return base64_encoded



from functools import partial
from typing import TYPE_CHECKING, Callable, Dict, Type, Union

import torch
from torch import nn

from eir.setup import schemas
from eir.setup.output_setup_modules.array_output_setup import ComputedArrayOutputInfo
from eir.setup.output_setup_modules.image_output_setup import ComputedImageOutputInfo
from eir.setup.output_setup_modules.sequence_output_setup import (
    ComputedSequenceOutputInfo,
)
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)
from eir.setup.schema_modules.output_schemas_tabular import TabularOutputTypeConfig
from eir.train_utils.metrics import calculate_prediction_losses, log_empty_loss_once

if TYPE_CHECKING:
    from eir.setup.output_setup import al_output_objects_as_dict
    from eir.setup.schema_modules.output_schemas_tabular import (
        al_cat_loss_names,
        al_con_loss_names,
    )

al_cat_losses = nn.CrossEntropyLoss
al_con_losses = (
    nn.MSELoss | nn.L1Loss | nn.SmoothL1Loss | nn.PoissonNLLLoss | nn.HuberLoss
)

al_criteria = al_con_losses | al_cat_losses | Callable

al_criteria_dict = Dict[str, Dict[str, al_criteria]]

al_losses = (
    nn.CrossEntropyLoss
    | nn.MSELoss
    | nn.L1Loss
    | nn.SmoothL1Loss
    | nn.PoissonNLLLoss
    | nn.HuberLoss
    | Callable
)


al_losses_classes = (
    Type[nn.CrossEntropyLoss]
    | Type[nn.MSELoss]
    | Type[nn.L1Loss]
    | Type[nn.SmoothL1Loss]
    | Type[nn.PoissonNLLLoss]
    | Type[nn.HuberLoss]
)


def get_criteria(outputs_as_dict: "al_output_objects_as_dict") -> al_criteria_dict:
    criteria_dict: al_criteria_dict = {}

    for output_name, output_object in outputs_as_dict.items():
        if output_name not in criteria_dict:
            criteria_dict[output_name] = {}

        match output_object:
            case ComputedTabularOutputInfo():
                target_col_iter = output_object.target_columns.items()
                for column_type, columns_of_type in target_col_iter:
                    for column_name in columns_of_type:
                        label_smoothing = _get_label_smoothing(
                            output_config=output_object.output_config,
                            column_type=column_type,
                        )

                        loss_name = _parse_loss_name(
                            output_config=output_object.output_config,
                            column_type=column_type,
                        )

                        criterion = get_supervised_criterion(
                            column_type_=column_type,
                            loss_name=loss_name,
                            cat_label_smoothing_=label_smoothing,
                        )

                        criteria_dict[output_name][column_name] = criterion

            case ComputedSequenceOutputInfo():
                pad_token = getattr(output_object.tokenizer, "pad_token", "<pad>")
                pad_idx = output_object.vocab[pad_token]

                criterion = partial(
                    _sequence_cat_loss,
                    cat_loss_func=nn.CrossEntropyLoss(ignore_index=pad_idx),
                )

                criteria_dict[output_name][output_name] = criterion

            case ComputedArrayOutputInfo() | ComputedImageOutputInfo():
                criterion = partial(_calc_con_loss, loss_func=nn.MSELoss())
                criteria_dict[output_name][output_name] = criterion

    return criteria_dict


def build_loss_dict() -> dict[str, list[str]]:
    loss_dict = {
        "cat": [
            "CrossEntropyLoss",
        ],
        "con": [
            "MSELoss",
            "L1Loss",
            "SmoothL1Loss",
            "PoissonNLLLoss",
            "HuberLoss",
        ],
    }

    return loss_dict


def get_supervised_criterion(
    column_type_: str, loss_name: str, cat_label_smoothing_: float = 0.0
) -> Union[nn.CrossEntropyLoss, Callable]:
    loss_dict = build_loss_dict()

    match column_type_, loss_name:
        case "con", _:
            assert loss_name in loss_dict["con"]

            loss_module = getattr(nn, loss_name)
            return partial(_calc_con_loss, loss_func=loss_module())

        case "cat", "CrossEntropyLoss":
            return nn.CrossEntropyLoss(label_smoothing=cat_label_smoothing_)

        case "cat", _:
            assert loss_name in loss_dict["cat"]
            loss_module = getattr(nn, loss_name)
            return loss_module()

    raise ValueError()


def _parse_loss_name(
    output_config: schemas.OutputConfig, column_type: str
) -> Union["al_cat_loss_names", "al_con_loss_names"]:
    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, TabularOutputTypeConfig)
    match column_type:
        case "cat":
            return output_type_info.cat_loss_name
        case "con":
            return output_type_info.con_loss_name
        case _:
            raise ValueError(f"Unknown column type: {column_type}")


def _get_label_smoothing(
    output_config: schemas.OutputConfig,
    column_type: str,
) -> float:
    output_type_info = output_config.output_type_info
    assert isinstance(output_type_info, TabularOutputTypeConfig)
    match column_type:
        case "con":
            return 0.0
        case "cat":
            return output_type_info.cat_label_smoothing
        case _:
            raise ValueError(f"Unknown column type: {column_type}")


def _calc_con_loss(
    input: torch.Tensor,
    target: torch.Tensor,
    loss_func: al_con_losses,
) -> torch.Tensor:
    match loss_func:
        case nn.PoissonNLLLoss():
            return loss_func(log_input=input.squeeze(), target=target.squeeze())
        case _:
            return loss_func(input=input.squeeze(), target=target.squeeze())


def get_loss_callable(criteria: al_criteria_dict) -> Callable:
    log_empty_callable = log_empty_loss_once()
    single_task_loss_func = partial(
        calculate_prediction_losses,
        criteria=criteria,
        log_empty_loss_callable=log_empty_callable,
    )
    return single_task_loss_func


def _sequence_cat_loss(
    input, target, cat_loss_func: nn.CrossEntropyLoss
) -> torch.Tensor:
    loss = cat_loss_func(input=input.transpose(2, 1), target=target)
    return loss



import reprlib
from pathlib import Path
from typing import TYPE_CHECKING, Sequence

import pandas as pd

from eir.setup import schemas
from eir.setup.schema_modules.output_schemas_array import ArrayOutputTypeConfig
from eir.setup.schema_modules.output_schemas_image import ImageOutputTypeConfig
from eir.setup.schema_modules.output_schemas_tabular import TabularOutputTypeConfig

if TYPE_CHECKING:
    from eir.setup.config import Configs


def validate_train_configs(configs: "Configs") -> None:
    validate_input_configs(input_configs=configs.input_configs)
    validate_output_configs(output_configs=configs.output_configs)

    validate_config_sync(
        input_configs=configs.input_configs, output_configs=configs.output_configs
    )

    validate_tensor_broker_configs(
        input_configs=configs.input_configs,
        fusion_configs=[configs.fusion_config],
        output_configs=configs.output_configs,
    )


def validate_input_configs(input_configs: Sequence[schemas.InputConfig]) -> None:
    for input_config in input_configs:

        input_source: str = input_config.input_info.input_source
        is_websocket = input_source.startswith("ws://")
        if is_websocket:
            continue

        input_source_path = Path(input_source)

        base_validate_input_info(input_info=input_config.input_info)

        match input_config.input_type_info:
            case schemas.TabularInputDataConfig(
                input_cat_columns, input_con_columns, _, _
            ):
                expected_columns = list(input_cat_columns) + list(input_con_columns)
                validate_tabular_source(
                    source_to_check=input_source_path,
                    expected_columns=expected_columns,
                    name="Tabular input",
                )


def base_validate_input_info(input_info: schemas.InputDataConfig) -> None:

    input_source = input_info.input_source

    if not Path(input_source).exists():
        raise ValueError(
            f"Input source {input_info.input_source} does not exist. "
            f"Please check the path is correct."
        )


def validate_output_configs(output_configs: Sequence[schemas.OutputConfig]) -> None:
    for output_config in output_configs:

        output_source = output_config.output_info.output_source
        is_websocket = output_source is not None and output_source.startswith("ws://")
        if is_websocket:
            continue

        base_validate_output_info(output_info=output_config.output_info)
        name = output_config.output_info.output_name

        match output_config.output_type_info:
            case TabularOutputTypeConfig(
                target_cat_columns,
                target_con_columns,
                _,
                _,
                _,
            ):
                output_source = output_config.output_info.output_source

                if output_source is None:
                    continue

                output_source_path = Path(output_source)

                expected_columns = list(target_cat_columns) + list(target_con_columns)
                validate_tabular_source(
                    source_to_check=output_source_path,
                    expected_columns=expected_columns,
                    name="Tabular output",
                )
            case ArrayOutputTypeConfig(_, _, loss, _):
                model_type = output_config.model_config.model_type
                if loss == "diffusion":
                    if model_type not in ("cnn",):
                        raise ValueError(
                            "Currently, diffusion loss is only supported for output "
                            "array model type 'cnn'. Please check the model type for "
                            f"output '{name}'."
                        )
            case ImageOutputTypeConfig(_, _, loss, _):
                model_type = output_config.model_config.model_type
                if loss == "diffusion":
                    if model_type not in ("cnn",):
                        raise ValueError(
                            "Currently, diffusion loss is only supported for output "
                            "image model type 'cnn'. Please check the model type for "
                            f"output '{name}'."
                        )


def base_validate_output_info(output_info: schemas.OutputInfoConfig) -> None:
    if output_info.output_source is None:
        return

    if not Path(output_info.output_source).exists():
        raise ValueError(
            f"Output source {output_info.output_source} does not exist. "
            f"Please check the path is correct."
        )


def validate_tabular_source(
    source_to_check: Path,
    expected_columns: Sequence[str],
    name: str,
) -> None:
    header = pd.read_csv(source_to_check, nrows=0).columns.tolist()
    if "ID" not in header:
        raise ValueError(
            f"{name} file {source_to_check} does not contain a column named 'ID'. "
            f"Please check the input file."
        )

    if not set(expected_columns).issubset(header):
        raise ValueError(
            f"{name} file {source_to_check} does not contain all expected columns: "
            f"{reprlib.repr(expected_columns)}. "
            f"Please check the input file."
        )

    series_ids = pd.read_csv(
        filepath_or_buffer=source_to_check,
        usecols=["ID"],
        engine="pyarrow",
        dtype_backend="pyarrow",
    ).squeeze()
    if not series_ids.is_unique:
        non_unique = series_ids[series_ids.duplicated()].tolist()
        raise ValueError(
            f"{name} file {source_to_check} contains non-unique"
            f" IDs: {reprlib.repr(non_unique)}. "
            f"Please check the input file."
        )


def validate_config_sync(
    input_configs: Sequence[schemas.InputConfig],
    output_configs: Sequence[schemas.OutputConfig],
) -> None:
    input_names = {input_config.input_info.input_name for input_config in input_configs}

    diff_out_configs = [
        i for i in output_configs if i.output_info.output_type in ("array", "image")
    ]
    for output_config in diff_out_configs:
        output_name = output_config.output_info.output_name
        output_type_info = output_config.output_type_info
        assert isinstance(
            output_type_info, (ArrayOutputTypeConfig, ImageOutputTypeConfig)
        )
        is_diffusion = output_type_info.loss == "diffusion"
        if is_diffusion and output_name not in input_names:
            raise ValueError(
                f"Diffusion loss is only supported when the corresponding input data "
                f"for the output '{output_name}' is provided. "
                "In practice, this means adding an input configuration with the "
                "same name and input_source as the output. "
                "This is needed to calculate the diffusion loss."
            )


def validate_tensor_broker_configs(
    input_configs: Sequence[schemas.InputConfig],
    fusion_configs: Sequence[schemas.FusionConfig],
    output_configs: Sequence[schemas.OutputConfig],
) -> None:
    all_tensor_broker_configs: list[schemas.TensorBrokerConfig] = []
    all_configs = list(input_configs) + list(fusion_configs) + list(output_configs)

    for config in all_configs:
        if (
            hasattr(config, "tensor_broker_config")
            and config.tensor_broker_config is not None
        ):
            all_tensor_broker_configs.append(config.tensor_broker_config)

    if not all_tensor_broker_configs:
        return

    all_message_names = []
    for broker_config in all_tensor_broker_configs:
        all_message_names.extend([msg.name for msg in broker_config.message_configs])

    if len(all_message_names) != len(set(all_message_names)):
        counts = {name: all_message_names.count(name) for name in all_message_names}
        raise ValueError(
            f"All tensor message names must be unique across all configs. "
            f"Got counts: {counts}"
        )

    for broker_config in all_tensor_broker_configs:
        for msg in broker_config.message_configs:
            if (
                msg.cache_fusion_type == "cross-attention"
                and msg.projection_type != "sequence"
            ):
                raise ValueError(
                    f"Cross-attention is only allowed with 'sequence' projection. "
                    f"Message '{msg.name}' uses {msg.projection_type}."
                )

    cached_messages = set()
    used_messages = set()
    for broker_config in all_tensor_broker_configs:
        for msg in broker_config.message_configs:
            if msg.cache_tensor:
                cached_messages.add(msg.name)
            if msg.use_from_cache:
                used_messages.update(msg.use_from_cache)

    unused_caches = cached_messages - used_messages
    if unused_caches:
        raise ValueError(
            f"The following tensor messages are never used: {unused_caches}"
        )

    unused_needing_cache = used_messages - cached_messages
    if unused_needing_cache:
        raise ValueError(
            f"The following tensor messages are used but not cached: "
            f"{unused_needing_cache}"
        )



import copy
import os
import sys
import warnings
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    DefaultDict,
    Dict,
    Generator,
    Iterable,
    Iterator,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Union,
)

# Filter warnings from attribution calculation
warnings.filterwarnings(
    "ignore", message="Setting forward, backward hooks and attributes on non-linear.*"
)
warnings.filterwarnings(
    "ignore", message=".*Attempting to normalize by value approximately 0.*"
)


import numpy as np
import torch
from aislib.misc_utils import ensure_path_exists
from captum.attr import NoiseTunnel
from ignite.engine import Engine
from torch import nn
from torch.cuda import OutOfMemoryError
from torch.utils.data import DataLoader, Subset
from torch.utils.hooks import RemovableHandle

from eir.data_load.data_utils import Batch, get_output_info_generator
from eir.data_load.datasets import al_local_datasets
from eir.interpretation.interpret_array import (
    ArrayConsumerCallable,
    analyze_array_input_attributions,
    get_array_sum_consumer,
)
from eir.interpretation.interpret_image import analyze_image_input_attributions
from eir.interpretation.interpret_omics import (
    OmicsConsumerCallable,
    analyze_omics_input_attributions,
    get_omics_consumer,
)
from eir.interpretation.interpret_sequence import analyze_sequence_input_attributions
from eir.interpretation.interpret_tabular import analyze_tabular_input_attributions
from eir.interpretation.interpretation_utils import MyIntegratedGradients
from eir.models.input.omics.omics_models import CNNModel, LinearModel
from eir.models.model_training_utils import gather_data_loader_samples
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)
from eir.setup.schemas import InputConfig, OutputConfig
from eir.target_setup.target_label_setup import MissingTargetsInfo
from eir.train_utils.evaluation import validation_handler
from eir.train_utils.utils import (
    call_hooks_stage_iterable,
    prepare_sample_output_folder,
    validate_handler_dependencies,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.data_load.label_setup import al_label_transformers_object
    from eir.models.model_setup_modules.meta_setup import al_meta_model
    from eir.predict import PredictExperiment
    from eir.predict_modules.predict_attributions import (
        LoadedTrainExperimentMixedWithPredict,
    )
    from eir.setup.input_setup import al_input_objects_as_dict
    from eir.train import Experiment
    from eir.train_utils.train_handlers import HandlerConfig

logger = get_logger(name=__name__, tqdm_compatible=True)

# Type aliases
al_explainers = Union[MyIntegratedGradients, NoiseTunnel]


class WrapperModelForAttribution(nn.Module):
    """
    We need this wrapper module because libraries often only handle torch.Tensor or
    List[torch.Tensor] inputs. However, we do not want to restrict our modules to
    only accept those formats, rather using a dict. Hence, we use this module to
    accept the list libraries expect, but call the wrapped model with a matched dict.
    """

    def __init__(
        self,
        wrapped_model,
        input_names: Iterable[str],
        output_name: str,
        column_name: str,
        *args,
        **kwargs,
    ):
        super().__init__()
        assert not wrapped_model.training
        self.wrapped_model = wrapped_model
        self.input_names = input_names
        self.output_name = output_name
        self.column_name = column_name

    def match_tuple_inputs_and_names(
        self, input_sequence: Sequence[torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        if isinstance(input_sequence, torch.Tensor):
            input_sequence = list(input_sequence)

        matched = {
            k: v
            for k, v in zip(self.input_names, input_sequence)
            if not k.startswith("__extras_")
        }

        return matched

    def forward(self, *inputs):
        """
        Note that have a slightly custom / different forward here compared with
        the method in MetaModel. This is because we only compute the attributions
        for tabular outputs, but we might have models that have e.g. diffusion
        outputs as well. In this case, the dynamic timestep embeddings do not
        contribute at all to the tabular output, which when computing the gradients
        in the Integrated Gradients method, will either lead to (a) differentiated
        tensors not being used in the graph and/or (b) None gradients showing up
        in the IG method.

        Hence, we manually only call the tabular output module we are currently
        interested in here, skipping e.g. the time embeddings completely.

        This also allows us to optimize slightly by only calling output modules
        that are actually needed.
        """
        inputs_matched = self.match_tuple_inputs_and_names(input_sequence=inputs)
        model = self.wrapped_model

        feature_extractors_out = {}
        for module_name, cur_input_module in model.input_modules.items():
            module_input = inputs_matched[module_name]
            feature_extractors_out[module_name] = cur_input_module(module_input)

        fused_features = {}
        for output_type, fusion_module in model.fusion_modules.items():
            fused_features[output_type] = fusion_module(feature_extractors_out)
        cur_fusion_target = model.fusion_to_output_mapping[self.output_name]
        fused_features = fused_features[cur_fusion_target]

        output_modules_out = {}
        cur_output = model.output_modules[self.output_name](fused_features)
        output_modules_out[self.output_name] = cur_output

        return output_modules_out[self.output_name][self.column_name]


@contextmanager
def suppress_stdout_and_stderr() -> Iterator[None]:
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        old_stderr = sys.stderr
        sys.stderr = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr


@validate_handler_dependencies([validation_handler])
def attribution_analysis_handler(
    engine: Engine, handler_config: "HandlerConfig"
) -> None:
    exp = handler_config.experiment
    gc = exp.configs.global_config
    iteration = engine.state.iteration

    types, should_run = check_if_should_run_analysis(
        input_configs=exp.configs.input_configs
    )
    if not should_run:
        logger.warning(
            "Got compute_attributions: %s but none of the input types in %s currently "
            "support attribution analysis. "
            "Attribution analysis will be skipped. ",
            gc.aa.compute_attributions,
            types,
        )
        return

    logger.debug("Running attribution analysis.")

    attribution_output_folder_callable = partial(
        _prepare_eval_attribution_outfolder,
        output_folder=gc.be.output_folder,
        iteration=iteration,
    )

    background_loader = get_background_loader(experiment=exp)

    tabular_attribution_analysis_wrapper(
        model=exp.model,
        experiment=exp,
        output_folder_target_callable=attribution_output_folder_callable,
        dataset_to_interpret=exp.valid_dataset,
        background_loader=background_loader,
    )


def check_if_should_run_analysis(
    input_configs: Sequence[InputConfig],
) -> Tuple[set, bool]:
    all_types = {i.input_info.input_type for i in input_configs}

    if all_types == {"bytes"}:
        return all_types, False
    return all_types, True


def get_background_loader(experiment: "Experiment") -> torch.utils.data.DataLoader:
    background_loader = copy.deepcopy(experiment.train_loader)

    return background_loader


def tabular_attribution_analysis_wrapper(
    model: "al_meta_model",
    experiment: Union[
        "Experiment", "PredictExperiment", "LoadedTrainExperimentMixedWithPredict"
    ],
    output_folder_target_callable: Callable,
    dataset_to_interpret: al_local_datasets,
    background_loader: torch.utils.data.DataLoader,
) -> None:
    """
    We need to copy the model to avoid affecting the actual model during
    training (e.g. zero-ing out gradients).
    """

    exp = experiment
    gc = experiment.configs.global_config

    model_copy = copy.deepcopy(model)
    model_copy.eval()
    target_columns_gen = get_output_info_generator(outputs_as_dict=exp.outputs)

    for output_name, target_column_type, target_column_name in target_columns_gen:
        output_type = exp.outputs[output_name].output_config.output_info.output_type

        if output_type != "tabular":
            continue

        ao = get_attribution_object(
            experiment=exp,
            model=model_copy,
            column_name=target_column_name,
            output_name=output_name,
            background_loader=background_loader,
            n_background_samples=gc.aa.attribution_background_samples,
        )

        act_callable = get_oom_adaptive_attribution_callable(
            explainer=ao.explainer,
            column_type=target_column_type,
            baseline_values=ao.baseline_values_ordered,
            baseline_names_ordered=ao.input_names_ordered,
            batch_size=gc.be.batch_size,
        )

        act_func = partial(
            get_attribution,
            attribution_callable=act_callable,
            input_names=ao.input_names_ordered,
        )

        data_producer = _get_interpretation_data_producer(
            experiment=exp,
            column_name=target_column_name,
            column_type=target_column_type,
            output_name=output_name,
            dataset=dataset_to_interpret,
        )

        act_producer = get_sample_attribution_producer(
            data_producer=data_producer,
            act_func=act_func,
            output_name=output_name,
            target_column_name=target_column_name,
            missing_ids_per_output=dataset_to_interpret.missing_ids_per_output,
        )

        input_names_and_types = _extract_input_names_and_types(
            input_objects=exp.inputs, input_names_ordered=ao.input_names_ordered
        )

        output_object = exp.outputs[output_name]
        assert isinstance(output_object, ComputedTabularOutputInfo)
        target_transformer = output_object.target_transformers[target_column_name]

        act_consumers = get_attribution_consumers(
            input_names_and_types=input_names_and_types,
            target_transformer=target_transformer,
            output_name=output_name,
            target_column=target_column_name,
            column_type=target_column_type,
        )

        all_attributions = process_attributions_for_all_modalities(
            attribution_producer=act_producer, attribution_consumers=act_consumers
        )

        for input_name in ao.input_names_ordered:
            input_object = exp.inputs[input_name]
            input_type = input_object.input_config.input_info.input_type

            if _do_skip_analyzing_input(
                input_name=input_name,
                input_type=input_type,
                output_configs=exp.configs.output_configs,
            ):
                continue

            act_output_folder = output_folder_target_callable(
                column_name=target_column_name,
                input_name=input_name,
                output_name=output_name,
            )
            common_kwargs = {
                "experiment": experiment,
                "input_name": input_name,
                "target_column_name": target_column_name,
                "target_column_type": target_column_type,
                "all_attributions": all_attributions[input_name],
                "attribution_outfolder": act_output_folder,
            }

            if input_type == "omics":
                analyze_omics_input_attributions(**common_kwargs)

            elif input_type == "tabular":
                analyze_tabular_input_attributions(
                    **common_kwargs, output_name=output_name
                )

            elif input_type == "sequence":
                # Currently we just run the full model here, despite being slightly
                # redundant, hence we pass in baselines_with_extras. Otherwise, we
                # would have to have a custom forward call skipping the time embeddings.
                expected_value = compute_expected_value(
                    model=model_copy,
                    baselines=ao.baselines_with_extras,
                )
                analyze_sequence_input_attributions(
                    **common_kwargs,
                    expected_target_classes_attributions=expected_value,
                    output_name=output_name,
                )
            elif input_type == "image":
                analyze_image_input_attributions(
                    **common_kwargs, output_name=output_name
                )

            elif input_type == "array":
                cur_array_attributions = all_attributions[input_name]
                assert isinstance(cur_array_attributions, dict)
                analyze_array_input_attributions(
                    attribution_outfolder=act_output_folder,
                    all_attributions=cur_array_attributions,
                )

        ao.hook_handle.remove()


def _extract_input_names_and_types(
    input_objects: "al_input_objects_as_dict",
    input_names_ordered: Sequence[str],
) -> dict[str, Literal["tabular", "omics", "sequence", "bytes", "image", "array"]]:
    input_names_and_types = {
        i: input_objects[i].input_config.input_info.input_type
        for i in input_names_ordered
        if not i.startswith("__extras_")
    }

    return input_names_and_types


def _do_skip_analyzing_input(
    input_name: str,
    input_type: str,
    output_configs: Sequence[OutputConfig],
) -> bool:
    """
    Second case is for when an input was generated for an output, e.g.
    in the case of sequence outputs.
    """
    if input_type == "bytes":
        return True

    input_name_in_output = any(
        input_name == output_config.output_info.output_name
        and output_config.output_info.output_type == "sequence"
        for output_config in output_configs
    )
    if input_name_in_output:
        return True

    return False


def compute_expected_value(
    model: nn.Module,
    baselines: Dict[str, torch.Tensor],
) -> np.ndarray:
    outputs = model(baselines)
    return outputs.mean(0).detach().cpu().numpy()


@dataclass
class AttributionObject:
    explainer: al_explainers
    hook_handle: RemovableHandle
    input_names_ordered: tuple[str, ...]
    baselines: dict[str, torch.Tensor]
    baseline_values_ordered: tuple[torch.Tensor, ...]
    baselines_with_extras: dict[str, torch.Tensor]


def get_attribution_object(
    experiment: Union[
        "Experiment",
        "PredictExperiment",
        "LoadedTrainExperimentMixedWithPredict",
    ],
    model: "al_meta_model",
    column_name: str,
    output_name: str,
    background_loader: DataLoader,
    n_background_samples: int,
) -> AttributionObject:
    """
    Note that currently we are always grabbing at least one batch of samples
    from the background loader (even when n_background_samples=0).
    """

    baseline_data, *_ = gather_data_loader_samples(
        batch_prep_hook=experiment.hooks.step_func_hooks.base_prepare_batch,
        batch_prep_hook_kwargs={"experiment": experiment},
        data_loader=background_loader,
        n_samples=n_background_samples,
    )

    baseline_data = _detach_all_inputs(tensor_inputs=baseline_data)
    baseline_data_no_extras = {
        k: v for k, v in baseline_data.items() if not k.startswith("__extras_")
    }

    hook_partial = partial(
        _grab_single_target_from_model_output_hook,
        output_target_column=column_name,
        output_name=output_name,
    )
    hook_handle = model.register_forward_hook(hook_partial)

    # Convert to list for wrapper model
    input_names, values_ordered = zip(*baseline_data_no_extras.items())
    input_names, values_ordered = tuple(input_names), tuple(values_ordered)

    assert not model.training
    wrapped_model = WrapperModelForAttribution(
        wrapped_model=model,
        input_names=input_names,
        output_name=output_name,
        column_name=column_name,
    )
    explainer = MyIntegratedGradients(forward_func=wrapped_model)

    attribution_object = AttributionObject(
        explainer=explainer,
        hook_handle=hook_handle,
        baselines=baseline_data_no_extras,
        input_names_ordered=input_names,
        baseline_values_ordered=values_ordered,
        baselines_with_extras=baseline_data,
    )

    return attribution_object


class AttributionCallable(Protocol):
    def __call__(
        self,
        *args,
        inputs: Dict[str, torch.Tensor],
        sample_label: torch.Tensor,
        **kwargs,
    ) -> list[np.ndarray]: ...


def get_attribution(
    attribution_callable: AttributionCallable,
    inputs: dict[str, torch.Tensor],
    input_names: Sequence[str],
    *args,
    **kwargs,
) -> Optional[dict[str, np.ndarray]]:
    input_attributions = attribution_callable(inputs=inputs, *args, **kwargs)

    if input_attributions is None:
        return None

    name_matched_attribution = {k: v for k, v in zip(input_names, input_attributions)}

    return name_matched_attribution


def get_oom_adaptive_attribution_callable(
    explainer: al_explainers,
    column_type: str,
    baseline_values: tuple[torch.Tensor, ...],
    baseline_names_ordered: tuple[str, ...],
    batch_size: int,
) -> AttributionCallable:
    n_steps = min(batch_size, 128)
    internal_batch_size = None
    has_successfully_run = False

    def calculate_attributions_adaptive(
        *args,
        inputs: Dict[str, torch.Tensor],
        sample_label: torch.Tensor,
        **kwargs,
    ) -> list[np.ndarray]:
        nonlocal has_successfully_run
        nonlocal internal_batch_size

        if has_successfully_run:
            return get_attributions(
                inputs=inputs,
                sample_label=sample_label,
                internal_batch_size=internal_batch_size,
                explainer=explainer,
                column_type=column_type,
                baselines=baseline_values,
                baseline_names_ordered=baseline_names_ordered,
                n_steps=n_steps,
            )
        else:
            while True:
                try:
                    res = get_attributions(
                        inputs=inputs,
                        sample_label=sample_label,
                        internal_batch_size=internal_batch_size,
                        explainer=explainer,
                        column_type=column_type,
                        baselines=baseline_values,
                        baseline_names_ordered=baseline_names_ordered,
                        n_steps=n_steps,
                    )
                    has_successfully_run = True

                    if internal_batch_size is not None:
                        logger.debug(
                            "Attribution IG internal batch size successfully set "
                            "to %d.",
                            internal_batch_size,
                        )

                    return res

                except OutOfMemoryError as e:  # type: ignore
                    if internal_batch_size == 1:
                        raise e

                    if internal_batch_size is None:
                        internal_batch_size = max(1, batch_size // 2)
                    else:
                        internal_batch_size = max(1, internal_batch_size // 2)

                    logger.debug(
                        "CUDA OOM error, reducing attribution IG "
                        "internal batch size to %d.",
                        internal_batch_size,
                    )
                    torch.cuda.empty_cache()
                    continue

    return calculate_attributions_adaptive


@dataclass
class SampleAttribution:
    sample_info: Batch
    sample_attributions: Dict[str, np.ndarray]
    raw_inputs: Dict


def get_attribution_consumers(
    input_names_and_types: dict[
        str, Literal["tabular", "omics", "sequence", "bytes", "image", "array"]
    ],
    target_transformer: "al_label_transformers_object",
    output_name: str,
    target_column: str,
    column_type: str,
) -> Dict[
    str, Union["BasicConsumerCallable", OmicsConsumerCallable, ArrayConsumerCallable]
]:
    consumers_dict = {}

    for input_name, input_type in input_names_and_types.items():
        if input_type == "bytes":
            continue

        consumer = _get_consumer_from_input_type(
            input_type=input_type,
            input_name=input_name,
            target_transformer=target_transformer,
            output_name=output_name,
            target_column=target_column,
            column_type=column_type,
        )
        if consumer is not None:
            consumers_dict[input_name] = consumer

    return consumers_dict


class BasicConsumerCallable(Protocol):
    def __call__(
        self,
        attribution: Optional["SampleAttribution"],
    ) -> Optional[Sequence["SampleAttribution"]]: ...


def _get_consumer_from_input_type(
    target_transformer: "al_label_transformers_object",
    input_type: str,
    input_name: str,
    output_name: str,
    target_column: str,
    column_type: str,
) -> BasicConsumerCallable | OmicsConsumerCallable | ArrayConsumerCallable:
    if input_type in ("sequence", "tabular", "image"):
        return get_basic_consumer()

    elif input_type == "omics":
        return get_omics_consumer(
            target_transformer=target_transformer,
            input_name=input_name,
            output_name=output_name,
            target_column=target_column,
            column_type=column_type,
        )

    elif input_type == "array":
        return get_array_sum_consumer(
            target_transformer=target_transformer,
            input_name=input_name,
            output_name=output_name,
            target_column=target_column,
            column_type=column_type,
        )
    else:
        raise ValueError(f"Unsupported input type {input_type}.")


def get_sample_attribution_producer(
    data_producer: Generator[tuple[Batch, dict[str, Any]], None, None],
    act_func: Callable,
    target_column_name: str,
    output_name: str,
    missing_ids_per_output: MissingTargetsInfo,
) -> Generator["SampleAttribution", None, None]:
    missing_for_modality = missing_ids_per_output.missing_ids_per_modality
    cur_missing_ids = missing_for_modality[output_name]

    missing_within_modality = missing_ids_per_output.missing_ids_within_modality
    cur_missing_ids_within = missing_within_modality[output_name][target_column_name]

    for batch, raw_inputs in data_producer:
        sample_target_labels = batch.target_labels
        cur_target_label = sample_target_labels[output_name][target_column_name]

        cur_id = batch.ids[0]
        if cur_id in cur_missing_ids or cur_id in cur_missing_ids_within:
            continue

        sample_all_modalities_attributions = act_func(
            inputs=batch.inputs,
            sample_label=cur_target_label,
        )

        if sample_all_modalities_attributions is None:
            continue

        batch_on_cpu = _convert_all_batch_tensors_to_cpu(batch=batch)
        cur_sample_attribution_info = SampleAttribution(
            sample_info=batch_on_cpu,
            sample_attributions=sample_all_modalities_attributions,
            raw_inputs=raw_inputs,
        )
        yield cur_sample_attribution_info


def get_basic_consumer() -> BasicConsumerCallable:
    results: list["SampleAttribution"] = []

    def _consumer(
        attribution: Optional["SampleAttribution"],
    ) -> Optional[Sequence["SampleAttribution"]]:
        if attribution is None:
            return results

        results.append(attribution)
        return None

    return _consumer


def process_attributions_for_all_modalities(
    attribution_producer: Generator["SampleAttribution", None, None],
    attribution_consumers: Dict[str, Callable],
) -> dict[str, Sequence["SampleAttribution"] | dict[str, np.ndarray]]:
    processed_attributions_all_modalities = {}

    for sample_attribution in attribution_producer:
        for consumer_name, consumer in attribution_consumers.items():
            parsed_act = _parse_out_non_target_modality_attributions(
                sample_attribution=sample_attribution, modality_key=consumer_name
            )
            consumer(parsed_act)

    for consumer_name, consumer in attribution_consumers.items():
        processed_attributions_all_modalities[consumer_name] = consumer(None)

    return processed_attributions_all_modalities


def _parse_out_non_target_modality_attributions(
    sample_attribution: SampleAttribution, modality_key: str
) -> SampleAttribution:
    parsed_act = {modality_key: sample_attribution.sample_attributions[modality_key]}
    parsed_inp = {modality_key: sample_attribution.raw_inputs[modality_key]}

    batch = sample_attribution.sample_info

    parsed_batch_inputs = {modality_key: batch.inputs[modality_key]}

    new_batch = Batch(
        inputs=parsed_batch_inputs, target_labels=batch.target_labels, ids=batch.ids
    )

    parsed_attribution = SampleAttribution(
        sample_info=new_batch, sample_attributions=parsed_act, raw_inputs=parsed_inp
    )

    return parsed_attribution


def _convert_all_batch_tensors_to_cpu(batch: Batch) -> Batch:
    """
    We need this to avoid blowing up GPU memory when gathering all the attributions and
    raw inputs, as the inputs are tensors which can be on the GPU.

    If needed later maybe we can use some fancy recursion here, but this works for now.
    """

    new_inputs = {k: v.cpu() for k, v in batch.inputs.items()}

    new_target_labels: Dict[str, Dict[str, torch.Tensor]] = {}
    for output_name, target_labels in batch.target_labels.items():
        target_labels_on_cpu = {k: v.cpu() for k, v in target_labels.items()}
        new_target_labels[output_name] = target_labels_on_cpu

    new_batch = Batch(inputs=new_inputs, target_labels=new_target_labels, ids=batch.ids)

    return new_batch


def _prepare_eval_attribution_outfolder(
    output_folder: str,
    input_name: str,
    column_name: str,
    output_name: str,
    iteration: int,
    *args,
    **kwargs,
):
    sample_outfolder = prepare_sample_output_folder(
        output_folder=output_folder,
        column_name=column_name,
        output_name=output_name,
        iteration=iteration,
    )
    attribution_outfolder = sample_outfolder / "attributions" / input_name
    ensure_path_exists(path=attribution_outfolder, is_folder=True)

    return attribution_outfolder


def _grab_single_target_from_model_output_hook(
    self: Union[CNNModel, LinearModel],
    input_: torch.Tensor,
    output: dict[str, dict[str, torch.Tensor]],
    output_target_column: str,
    output_name: str,
) -> torch.Tensor:
    return output[output_name][output_target_column]


def get_attributions(
    explainer: al_explainers,
    inputs: Dict[str, torch.Tensor],
    sample_label: torch.Tensor,
    column_type: str,
    baselines: tuple[torch.Tensor, ...],
    baseline_names_ordered: tuple[str, ...],
    n_steps: int,
    internal_batch_size: Union[int, None],
) -> list[np.ndarray]:

    tuple_inputs = tuple(inputs[n] for n in baseline_names_ordered)
    baselines_averaged = tuple(i.mean(0).unsqueeze(0) for i in baselines)

    common_kwargs = dict(
        inputs=tuple_inputs,
        baselines=baselines_averaged,
        n_steps=n_steps,
        internal_batch_size=internal_batch_size,
    )

    if column_type == "con":
        output = explainer.attribute(**common_kwargs)
    else:
        assert sample_label >= 0, f"Got {sample_label} for {inputs} and {column_type}."

        output = explainer.attribute(target=sample_label, **common_kwargs)

    output = [o.detach().cpu().numpy() for o in output]

    if column_type == "con":
        assert isinstance(output[0], np.ndarray)
        return output

    return output


def _get_interpretation_data_producer(
    experiment: Union[
        "Experiment", "PredictExperiment", "LoadedTrainExperimentMixedWithPredict"
    ],
    column_name: str,
    column_type: str,
    output_name: str,
    dataset: al_local_datasets,
) -> Generator[Tuple[Batch, dict[str, Any]], None, None]:
    output_object = experiment.outputs[output_name]
    assert isinstance(output_object, ComputedTabularOutputInfo)

    target_transformer = output_object.target_transformers[column_name]
    gc = experiment.configs.global_config

    target_classes_numerical = _get_numerical_target_classes(
        target_transformer=target_transformer,
        column_type=column_type,
    )

    attributions_data_loader = _get_attributions_dataloader(
        dataset=dataset,
        max_attributions_per_class=gc.aa.max_attributions_per_class,
        output_name=output_name,
        target_column=column_name,
        column_type=column_type,
        target_classes_numerical=target_classes_numerical,
    )

    for loader_batch in attributions_data_loader:
        state = call_hooks_stage_iterable(
            hook_iterable=experiment.hooks.step_func_hooks.base_prepare_batch,
            common_kwargs={"experiment": experiment, "loader_batch": loader_batch},
            state=None,
        )

        batch = state["batch"]

        inputs_detached = _detach_all_inputs(tensor_inputs=batch.inputs)
        batch_interpretation = Batch(
            inputs=inputs_detached,
            target_labels=batch.target_labels,
            ids=batch.ids,
        )

        raw_inputs = {k: v for k, v in loader_batch[0].items()}

        yield batch_interpretation, raw_inputs


def _detach_all_inputs(tensor_inputs: Dict[str, torch.Tensor]):
    inputs_detached = {}

    for input_name, value in tensor_inputs.items():
        inputs_detached[input_name] = value.detach()

    return inputs_detached


def _get_attributions_dataloader(
    dataset: al_local_datasets,
    max_attributions_per_class: Optional[int],
    output_name: str,
    target_column: str,
    column_type: str,
    target_classes_numerical: Sequence[int],
) -> DataLoader:
    if max_attributions_per_class is None:
        data_loader = DataLoader(dataset=dataset, batch_size=1, shuffle=False)
        return data_loader

    indices_func = _get_categorical_sample_indices_for_attributions
    if column_type == "con":
        indices_func = _get_continuous_sample_indices_for_attributions

    subset_indices = indices_func(
        dataset=dataset,
        max_attributions_per_class=max_attributions_per_class,
        target_column=target_column,
        output_name=output_name,
        target_classes_numerical=target_classes_numerical,
    )
    subset_dataset = _subsample_dataset(dataset=dataset, indices=subset_indices)
    data_loader = DataLoader(dataset=subset_dataset, batch_size=1, shuffle=False)
    return data_loader


def _get_categorical_sample_indices_for_attributions(
    dataset: al_local_datasets,
    max_attributions_per_class: int,
    output_name: str,
    target_column: str,
    target_classes_numerical: Sequence[int],
) -> Tuple[int, ...]:
    acc_label_counts: DefaultDict[int | float | torch.Tensor, int] = defaultdict(
        lambda: 0
    )
    acc_label_limit = max_attributions_per_class
    indices = []

    for index, sample in enumerate(dataset.samples):
        target_labels = sample.target_labels
        cur_sample_target_label = target_labels[output_name][target_column]

        is_over_limit = acc_label_counts[cur_sample_target_label] == acc_label_limit
        is_not_in_target_classes = (
            cur_sample_target_label not in target_classes_numerical
        )
        if is_over_limit or is_not_in_target_classes:
            continue

        indices.append(index)
        acc_label_counts[cur_sample_target_label] += 1

    return tuple(indices)


def _get_continuous_sample_indices_for_attributions(
    dataset: al_local_datasets, max_attributions_per_class: int, *args, **kwargs
) -> Tuple[int, ...]:
    acc_label_limit = max_attributions_per_class
    num_sample = len(dataset)
    indices = np.random.choice(num_sample, acc_label_limit)

    return tuple(indices)


def _subsample_dataset(dataset: al_local_datasets, indices: Sequence[int]):
    dataset_subset = Subset(dataset=dataset, indices=indices)
    return dataset_subset


def _get_numerical_target_classes(
    target_transformer,
    column_type: str,
):
    if column_type == "con":
        return [None]

    return target_transformer.transform(target_transformer.classes_)



import reprlib
import typing
from collections import OrderedDict
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Sequence, Tuple, Type, Union

import torch

from eir.models.model_setup_modules.meta_setup import al_meta_model
from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


def load_model(
    model_path: Path,
    model_class: Type[al_meta_model],
    model_init_kwargs: Dict,
    device: str,
    test_mode: bool,
    state_dict_keys_to_keep: Union[None, Sequence[str]] = None,
    state_dict_key_rename: Union[None, Sequence[Tuple[str, str]]] = None,
    strict_shapes: bool = True,
) -> al_meta_model:
    model = model_class(**model_init_kwargs)

    model = _load_model_weights(
        model=model,
        model_state_dict_path=model_path,
        device=device,
        state_dict_keys_to_keep=state_dict_keys_to_keep,
        state_dict_key_rename=state_dict_key_rename,
        strict_shapes=strict_shapes,
    )

    if test_mode:
        model.eval()

    return model


def _load_model_weights(
    model: al_meta_model,
    model_state_dict_path: Path,
    device: str,
    state_dict_keys_to_keep: Union[None, Sequence[str]] = None,
    state_dict_key_rename: Union[None, Sequence[Tuple[str, str]]] = None,
    strict_shapes: bool = True,
) -> al_meta_model:
    loaded_weights_state_dict = torch.load(
        model_state_dict_path,
        map_location=device,
        weights_only=True,
    )

    if state_dict_keys_to_keep:
        no_keys_before = len(loaded_weights_state_dict)
        loaded_weights_state_dict = _filter_state_dict_keys(
            state_dict=loaded_weights_state_dict, keys_to_keep=state_dict_keys_to_keep
        )
        logger.info(
            "Extracting %d/%d modules for feature extractors: '%s' from %s.",
            len(loaded_weights_state_dict),
            no_keys_before,
            state_dict_keys_to_keep,
            model_state_dict_path,
        )

    if state_dict_key_rename:
        for replace_tuple in state_dict_key_rename:
            logger.debug(
                "Renaming '%s' in pretrained model to '%s' in current model.",
                replace_tuple[0],
                replace_tuple[1],
            )
            loaded_weights_state_dict = _replace_dict_key_names(
                dict_=loaded_weights_state_dict,
                replace_pattern=replace_tuple,
            )

    if not strict_shapes:
        model_state_dict = model.state_dict()
        loaded_weights_state_dict = _filter_incompatible_parameter_shapes_for_loading(
            source_state_dict=model_state_dict,
            destination_state_dict=loaded_weights_state_dict,
        )

    incompatible_keys = model.load_state_dict(
        state_dict=loaded_weights_state_dict,
        strict=False,
    )

    total_keys = len(loaded_weights_state_dict)
    no_missing = len(incompatible_keys.missing_keys)
    no_unexpected = len(incompatible_keys.unexpected_keys)
    no_incompatible_keys = no_missing + no_unexpected
    if no_incompatible_keys > 0:
        repr_object = reprlib.Repr()
        repr_object.maxother = 256
        repr_object.maxstring = 256
        logger.info(
            "Encountered incompatible modules when loading model from '%s'.\n"
            "Missing keys: \n%s\n"
            "Unexpected keys: \n%s\n"
            "This is expected if you are loading select modules from a saved model, "
            "which means you can ignore this message. If you are loading a pre-trained "
            "model as-is, then this is most likely an error and something unexpected "
            "has changed between the pre-training and setting up the current model "
            "from the pre-trained one.",
            model_state_dict_path,
            repr_object.repr(incompatible_keys.missing_keys),
            repr_object.repr(incompatible_keys.unexpected_keys),
        )

    num_loaded = total_keys - no_incompatible_keys
    logger.info(
        "Successfully loaded %d/%d modules from %s.",
        num_loaded,
        total_keys,
        model_state_dict_path,
    )

    torch_device = torch.device(device)
    model = model.to(device=torch_device)

    return model


def _replace_dict_key_names(
    dict_: Dict[str, Any], replace_pattern: Tuple[str, str]
) -> OrderedDict:
    renamed_dict = OrderedDict()

    for key, value in dict_.items():
        new_key = key.replace(*replace_pattern)
        renamed_dict[new_key] = value

    return renamed_dict


def _filter_state_dict_keys(
    state_dict: typing.OrderedDict[str, torch.nn.Parameter], keys_to_keep: Sequence[str]
) -> typing.OrderedDict[str, torch.nn.Parameter]:
    filtered_state_dict = OrderedDict()

    for module_name, module_parameter in state_dict.items():
        if any(key in module_name for key in keys_to_keep):
            filtered_state_dict[module_name] = module_parameter

    return filtered_state_dict


def _filter_incompatible_parameter_shapes_for_loading(
    source_state_dict: Dict[str, Any], destination_state_dict: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Note that the way this is used is a bit unclear / backwards, but the destination
    state dict is the one that contains weights loaded from the pretrained model,
    and the source state dict is the current, randomly initialized model.

    When there is a shape that is incompatible, but with the same name, we replace
    the parameter in the destination state dict with the one from the source state.
    This ensures that when we model.load_state_dict(...), we effectively maintain
    the original weights in the source model, vs. alternatively having a clash
    trying to load incompatible weights into the source model that happen to have
    the same name.
    """

    destination_state_dict = deepcopy(destination_state_dict)

    for key in destination_state_dict:
        if key in source_state_dict:
            if destination_state_dict[key].shape != source_state_dict[key].shape:
                logger.info(
                    f"Skipping loading of parameter: {key} "
                    f"due to incompatible shapes. "
                    f"Source shape: {source_state_dict[key].shape}. "
                    f"Destination shape: {destination_state_dict[key].shape}."
                )
                destination_state_dict[key] = source_state_dict[key]

    return destination_state_dict



from functools import partial
from typing import TYPE_CHECKING, Callable, Dict, Literal, NewType, Type, cast

import torch
from torch import nn

from eir.models.fusion import fusion_default, fusion_identity, fusion_mgmoe
from eir.models.layers.mlp_layers import ResidualMLPConfig
from eir.models.meta.meta_utils import FusionModuleProtocol, al_fusion_modules
from eir.utils.logging import get_logger

al_fusion_model = Literal["pass-through", "mlp-residual", "identity", "mgmoe"]
al_fusion_model_configs = (
    ResidualMLPConfig | fusion_identity.IdentityConfig | fusion_mgmoe.MGMoEModelConfig
)

ComputedType = NewType("ComputedType", torch.Tensor)
PassThroughType = NewType("PassThroughType", Dict[str, torch.Tensor])
al_fused_features = dict[str, ComputedType | PassThroughType | torch.Tensor]

if TYPE_CHECKING:
    from eir.models.meta.meta_utils import al_input_modules

logger = get_logger(name=__name__)


def pass_through_fuse(features: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
    return features


def get_fusion_modules(
    fusion_model_type: str,
    model_config: al_fusion_model_configs,
    modules_to_fuse: "al_input_modules",
    out_feature_per_feature_extractor: Dict[str, int],
    output_types: dict[str, Literal["tabular", "sequence", "array"]],
    any_diffusion: bool,
    strict: bool = True,
) -> al_fusion_modules:
    if strict:
        _check_fusion_modules(
            output_types=output_types,
            fusion_model_type=fusion_model_type,
        )

    fusion_modules: al_fusion_modules = cast(al_fusion_modules, nn.ModuleDict())

    fusion_in_dim = _get_fusion_input_dimension(modules_to_fuse=modules_to_fuse)
    any_tabular = any(i for i in output_types.values() if i in ("tabular",))
    any_sequence = any(i for i in output_types.values() if i in ("sequence",))

    any_array = any(i for i in output_types.values() if i in ("array", "image"))
    array_and_no_diffusion = any_array and not any_diffusion
    array_and_diffusion = any_array and any_diffusion

    if any_tabular or array_and_no_diffusion:
        fusion_class = get_fusion_class(fusion_model_type=fusion_model_type)
        computing_fusion_module = fusion_class(
            model_config=model_config,
            fusion_in_dim=fusion_in_dim,
            out_feature_per_feature_extractor=out_feature_per_feature_extractor,
        )
        fusion_modules["computed"] = computing_fusion_module

    if any_sequence or array_and_diffusion:
        model_config = fusion_identity.IdentityConfig()
        pass_through_fusion_module = fusion_identity.IdentityFusionModel(
            model_config=model_config,
            fusion_in_dim=fusion_in_dim,
            fusion_callable=pass_through_fuse,
        )
        fusion_modules["pass-through"] = cast(
            FusionModuleProtocol, pass_through_fusion_module
        )

    assert len(fusion_modules) > 0

    return fusion_modules


def _check_fusion_modules(
    output_types: dict[str, Literal["tabular", "sequence", "array"]],
    fusion_model_type: str,
) -> None:
    """
    Note we skip the 'array' here as it can be both computed and pass-through.
    """
    if not output_types:
        raise ValueError("output_types cannot be empty.")

    output_set = set(output_types.values())
    computed_set = {
        "tabular",
    }
    pass_through_set = {
        "sequence",
    }
    full_set = computed_set.union(pass_through_set).union({"array"})
    supported_fusion_models = {
        "mlp-residual",
        "mgmoe",
        "identity",
        "pass-through",
    }

    if not full_set:
        raise ValueError(
            f"Invalid output type(s). "
            f"Supported types are {computed_set.union(pass_through_set)}."
        )

    if fusion_model_type not in supported_fusion_models:
        raise ValueError(
            f"Invalid fusion_model_type. Supported types are {supported_fusion_models}."
        )

    if output_set == pass_through_set and fusion_model_type != "pass-through":
        raise ValueError(
            f"When using only {pass_through_set} outputs, "
            f"only pass-through is supported. "
            f"Got {fusion_model_type}. To use pass-through, "
            f"set fusion_model_type to 'pass-through'."
        )

    elif output_set.issubset(computed_set) and fusion_model_type == "pass-through":
        raise ValueError(
            f"When using only {computed_set} outputs, pass-through is not supported. "
            f"Got {fusion_model_type}. Kindly set the fusion_model_type "
            "to 'mlp-residual', 'mgmoe', or 'identity'."
        )

    elif (
        pass_through_set.intersection(output_set)
        and fusion_model_type != "pass-through"
    ):
        logger.warning(
            f"Note: When using {output_set} outputs, "
            f"the fusion model type {fusion_model_type} "
            f"is only applied to the {computed_set} outputs. "
            f"The fusion for {pass_through_set} outputs "
            "is handled by the respective output module themselves, "
            "and the feature representations are passed through"
            " directly to the output module."
        )


def _get_fusion_input_dimension(modules_to_fuse: "al_input_modules") -> int:
    fusion_in_dim = sum(i.num_out_features for i in modules_to_fuse.values())
    return fusion_in_dim


def get_fusion_class(
    fusion_model_type: str,
) -> Type[nn.Module] | Callable:
    if fusion_model_type == "mgmoe":
        return fusion_mgmoe.MGMoEModel
    elif fusion_model_type == "mlp-residual":
        return fusion_default.MLPResidualFusionModule
    elif fusion_model_type == "identity":
        return fusion_identity.IdentityFusionModel
    elif fusion_model_type == "pass-through":
        return partial(
            fusion_identity.IdentityFusionModel, fusion_callable=pass_through_fuse
        )
    raise ValueError(f"Unrecognized fusion model type: {fusion_model_type}.")



import torch

from eir.models.model_setup_modules.output_model_setup_modules import al_output_modules
from eir.models.output.tabular.linear import (
    LinearOutputModule,
    LinearOutputModuleConfig,
)
from eir.models.output.tabular.mlp_residual import (
    ResidualMLPOutputModule,
    ResidualMLPOutputModuleConfig,
)
from eir.models.output.tabular.tabular_output_modules import TabularOutputModuleConfig
from eir.setup.output_setup_modules.tabular_output_setup import (
    al_num_outputs_per_target,
)


def get_tabular_output_module_from_model_config(
    output_model_config: TabularOutputModuleConfig,
    input_dimension: int,
    num_outputs_per_target: "al_num_outputs_per_target",
    device: str,
) -> al_output_modules:
    output_module_type = output_model_config.model_type
    model_init_config = output_model_config.model_init_config

    output_module: al_output_modules
    match output_module_type:
        case "mlp_residual":
            assert isinstance(model_init_config, ResidualMLPOutputModuleConfig)
            output_module = ResidualMLPOutputModule(
                model_config=model_init_config,
                input_dimension=input_dimension,
                num_outputs_per_target=num_outputs_per_target,
            )
        case "linear":
            assert isinstance(model_init_config, LinearOutputModuleConfig)
            output_module = LinearOutputModule(
                model_config=model_init_config,
                input_dimension=input_dimension,
                num_outputs_per_target=num_outputs_per_target,
            )
        case _:
            raise ValueError(f"Invalid output module type: {output_module_type}")

    torch_device = torch.device(device=device)
    output_module = output_module.to(device=torch_device)

    return output_module



from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Literal

import dill
import torch
from torch import nn
from torch.nn import functional as F

from eir.models.fusion.fusion_attention import MetaSequenceProjection
from eir.models.input.sequence.transformer_models import (
    BasicTransformerFeatureExtractorModelConfig,
    TransformerWrapperModel,
    parse_dim_feedforward,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.models.model_setup_modules.meta_setup import (
        FeatureExtractorInfo,
        al_meta_model,
    )
    from eir.setup.input_setup import al_input_objects_as_dict
    from eir.setup.output_setup_modules.sequence_output_setup import (
        ComputedSequenceOutputInfo,
    )

al_sequence_output_models = Literal["sequence"]

logger = get_logger(name=__name__)


@dataclass
class TransformerSequenceOutputModuleConfig(
    BasicTransformerFeatureExtractorModelConfig
):
    pass


@dataclass
class SequenceOutputModuleConfig:
    """
    :param model_init_config:
          Configuration / arguments used to initialise model.

    :param model_type:
         Which type of image model to use.

    :param embedding_dim:
        Which dimension to use for the embeddings. If ``None``, will automatically set
        this value based on the number of tokens and attention heads.

    :param position:
        Whether to encode the token position or use learnable position embeddings.

    :param position_dropout:
        Dropout for the positional encoding / embedding.

    """

    model_init_config: TransformerSequenceOutputModuleConfig
    model_type: Literal["sequence"] = "sequence"

    embedding_dim: int = 64

    position: Literal["encode", "embed"] = "encode"
    position_dropout: float = 0.10

    projection_layer_type: Literal["auto", "lcl", "lcl_residual", "linear"] = "auto"


class SequenceOutputModule(nn.Module):
    def __init__(
        self,
        output_object: "ComputedSequenceOutputInfo",
        output_name: str,
        feature_dimensionalities_and_types: Dict[str, "FeatureExtractorInfo"],
        *args,
        **kwargs,
    ):
        super().__init__()

        self.num_tokens = len(output_object.vocab)
        self.input_dimensions = feature_dimensionalities_and_types
        self.embedding_dim = output_object.embedding_dim
        self.max_length = output_object.computed_max_length
        self.output_name = output_name
        self.output_model_config = output_object.output_config.model_config
        assert isinstance(self.output_model_config, SequenceOutputModuleConfig)
        self.output_model_init_config = self.output_model_config.model_init_config
        assert isinstance(
            self.output_model_init_config, TransformerSequenceOutputModuleConfig
        )

        dim_feed_forward = parse_dim_feedforward(
            dim_feedforward=self.output_model_init_config.dim_feedforward,
            embedding_dim=self.embedding_dim,
        )

        transformer_output_layer_base = nn.TransformerEncoderLayer(
            d_model=self.embedding_dim,
            nhead=self.output_model_init_config.num_heads,
            dim_feedforward=dim_feed_forward,
            dropout=self.output_model_init_config.dropout,
            activation="gelu",
            batch_first=True,
            norm_first=True,
        )

        mask = torch.triu(
            torch.ones(self.max_length, self.max_length) * float("-inf"),
            diagonal=1,
        )
        self.register_buffer("mask", mask)

        self.output_transformer = nn.TransformerEncoder(
            encoder_layer=transformer_output_layer_base,
            num_layers=self.output_model_init_config.num_layers,
            enable_nested_tensor=False,
        )

        self.match_projections = nn.ModuleDict()
        for input_name, feature_extractor_info in self.input_dimensions.items():
            if input_name == self.output_name:
                continue

            match feature_extractor_info.input_type:
                case "sequence":
                    in_embed = feature_extractor_info.input_dimension.width
                case _:
                    in_embed = feature_extractor_info.output_dimension

            in_elements = feature_extractor_info.output_dimension
            cur_projection = MetaSequenceProjection(
                in_total_num_elements=in_elements,
                in_embedding_dim=in_embed,
                target_embedding_dim=self.embedding_dim,
                target_max_length=self.max_length,
                projection_layer_type=self.output_model_config.projection_layer_type,
            )

            self.match_projections[input_name] = cur_projection

        self.head = nn.Linear(self.embedding_dim, self.num_tokens)

    def forward(self, input: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        out = input[self.output_name]
        out = out.reshape(out.shape[0], self.max_length, self.embedding_dim)

        for input_name, input_tensor in input.items():
            if input_name == self.output_name:
                continue

            cur_projection = self.match_projections[input_name]
            projected = cur_projection(input_tensor=input_tensor, target_tensor=out)
            out = out + projected

        out = self.output_transformer(out, mask=self.mask)

        out = self.head(out)

        return {self.output_name: out}


def overload_embeddings_with_pretrained(
    model: "al_meta_model",
    inputs: "al_input_objects_as_dict",
    pretrained_checkpoint: str,
) -> "al_meta_model":
    """
    Vocab: From serialized input object
    Embeddings: From loaded model

    - If we have a pretrained checkpoint in global config, we have to initialize inputs
    from that experiment in order to get the vocab per input.

    - If we are using selected ones, we can just use the vocab from the input object
    here directly.

    In both cases, we have to load the pretrained model to grab the embeddings.
    Probably it's enough to just use the torch.load() function here, since it's just
    giving us a dictionary.

    First, let's just assume the global case.
    """

    if not pretrained_checkpoint:
        return model

    any_sequence_inputs = any(
        input_object.input_config.input_info.input_type == "sequence"
        for input_object in inputs.values()
    )
    if not any_sequence_inputs:
        return model

    logger.info(
        f"Overloading embeddings with pretrained checkpoint {pretrained_checkpoint}."
    )

    run_folder = Path(pretrained_checkpoint).parent.parent
    serialized_inputs = run_folder / "serializations/sequence_input_serializations"

    any_serialized_sequence_inputs = serialized_inputs.exists()
    if not any_serialized_sequence_inputs:
        return model

    input_objects_loaded = {}
    for serialized_input in serialized_inputs.iterdir():
        input_name = serialized_input.stem
        with open(serialized_input, "rb") as f:
            input_object = dill.load(file=f)
        input_objects_loaded[input_name] = input_object

    loaded_state_dict = torch.load(
        f=pretrained_checkpoint,
        weights_only=True,
    )

    for input_name, input_object in inputs.items():
        input_type = input_object.input_config.input_info.input_type
        if input_type != "sequence":
            continue

        cur_vocab = input_object.vocab.get_stoi()
        prev_input_object_vocab = input_objects_loaded[input_name].vocab.get_stoi()

        prev_emb_key = f"input_modules.{input_name}.embedding.weight"
        prev_embeddings = loaded_state_dict[prev_emb_key]

        cur_input_module = model.input_modules[input_name]
        assert isinstance(cur_input_module, TransformerWrapperModel)

        cur_embedding = cur_input_module.embedding.weight
        cur_embedding_copy = cur_embedding.clone().detach()

        for token, idx in cur_vocab.items():
            if token not in prev_input_object_vocab:
                continue

            prev_idx = prev_input_object_vocab[token]
            prev_emb = prev_embeddings[prev_idx]

            cur_emb = cur_embedding_copy[idx]

            if prev_emb.shape != cur_emb.shape:
                logger.warning(
                    f"Shape mismatch for token {token} in input {input_name}."
                    f"Applying average pooling to match dimensions."
                )
                prev_emb = prev_emb.view(1, 1, -1)
                prev_emb = F.adaptive_avg_pool1d(
                    input=prev_emb,
                    output_size=cur_emb.shape[0],
                )
                prev_emb = prev_emb.view(-1)

            cur_embedding_copy[idx] = prev_emb

        cur_input_module.embedding.weight = nn.Parameter(
            data=cur_embedding_copy,
            requires_grad=True,
        )
        logger.info(f"Overloaded embeddings for {input_name}.")

    return model



from dataclasses import dataclass, field
from typing import List

import torch
from torch import nn
from torchvision.ops import StochasticDepth

from eir.models.layers.norm_layers import LayerScale


@dataclass
class ResidualMLPConfig:
    """
    :param layers:
        Number of residual MLP layers to use in for each output predictor after fusing.

    :param fc_task_dim:
        Number of hidden nodes in each MLP residual block.

    :param rb_do:
        Dropout in each MLP residual block.

    :param fc_do:
        Dropout before final layer.

    :param stochastic_depth_p:
        Probability of dropping input.

    """

    layers: List[int] = field(default_factory=lambda: [2])

    fc_task_dim: int = 256

    rb_do: float = 0.10
    fc_do: float = 0.10

    stochastic_depth_p: float = 0.10


class MLPResidualBlock(nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        dropout_p: float = 0.0,
        full_preactivation: bool = False,
        stochastic_depth_p: float = 0.0,
    ):
        super().__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.dropout_p = dropout_p
        self.full_preactivation = full_preactivation
        self.stochastic_depth_p = stochastic_depth_p

        self.norm_1 = nn.LayerNorm(normalized_shape=in_features)

        self.fc_1 = nn.Linear(
            in_features=in_features, out_features=out_features, bias=True
        )

        self.act_1 = nn.GELU()
        self.do = nn.Dropout(p=dropout_p)
        self.fc_2 = nn.Linear(
            in_features=out_features, out_features=out_features, bias=True
        )

        if in_features == out_features:
            self.downsample_identity = lambda x: x
        else:
            self.downsample_identity = nn.Linear(
                in_features=in_features,
                out_features=out_features,
                bias=True,
            )

        self.ls = LayerScale(dim=out_features, init_values=1e-05)

        self.stochastic_depth = StochasticDepth(p=self.stochastic_depth_p, mode="batch")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.norm_1(x)

        identity = out if self.full_preactivation else x
        identity = self.downsample_identity(identity)

        out = self.fc_1(out)

        out = self.act_1(out)
        out = self.do(out)
        out = self.fc_2(out)
        out = self.ls(out)

        out = self.stochastic_depth(out)

        return out + identity



from pathlib import Path
from typing import TYPE_CHECKING, Sequence, Tuple, Union

from eir.experiment_io.experiment_io import (
    get_run_folder_from_model_path,
    load_serialized_train_experiment,
)
from eir.models.model_setup_modules.meta_setup import (
    MetaClassGetterCallable,
    get_default_meta_class,
    get_meta_model_class_and_kwargs_from_configs,
)
from eir.models.model_setup_modules.model_io import load_model
from eir.setup import schemas
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.output_setup import al_output_objects_as_dict
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.models.meta.meta_utils import al_input_modules

logger = get_logger(name=__name__)


def overload_meta_model_feature_extractors_with_pretrained(
    input_modules: "al_input_modules",
    inputs_as_dict: al_input_objects_as_dict,
    outputs_as_dict: "al_output_objects_as_dict",
    meta_class_getter: MetaClassGetterCallable = get_default_meta_class,
) -> "al_input_modules":
    """
    Note that `inputs_as_dict` here are coming from the current experiment, arguably
    it would be more robust / better to have them loaded from the pretrained experiment,
    but then we have to setup things from there such as hooks, valid_ids, train_ids,
    etc.

    For now, we will enforce that the feature extractor architecture that is set-up
    and then uses pre-trained weights from a previous experiment must match that of
    the feature extractor that did the pre-training. Simply put, we must ensure
    that all input setup parameters that have to do with architecture match exactly
    between the (a) pretrained input config and (b) the input config loading the
    pretrained model.

    The strict parameter being set to false is due to the fact that we can have e.g.
    different fusion modules in the current model vs the pretrained model, which should
    raise an error e.g. if continuing training from a checkpoint. However in this case,
    it acceptable to have different fusion modules, as long as the input modules are
    matching.
    """

    any_pretrained = any(
        i.input_config.pretrained_config for i in inputs_as_dict.values()
    )
    if not any_pretrained:
        return input_modules

    input_configs = tuple(i.input_config for i in inputs_as_dict.values())
    replace_pattern = _build_all_replacements_tuples_for_loading_pretrained_module(
        input_configs=input_configs
    )
    for input_name, input_object in inputs_as_dict.items():
        input_config = input_object.input_config

        pretrained_config = input_config.pretrained_config
        if not pretrained_config:
            continue

        load_model_path = Path(pretrained_config.model_path)
        load_run_folder = get_run_folder_from_model_path(
            model_path=str(load_model_path)
        )
        load_experiment = load_serialized_train_experiment(run_folder=load_run_folder)
        load_configs = load_experiment.configs

        func = get_meta_model_class_and_kwargs_from_configs
        meta_model_class, meta_model_kwargs = func(
            global_config=load_configs.global_config,
            fusion_config=load_configs.fusion_config,
            inputs_as_dict=inputs_as_dict,
            outputs_as_dict=outputs_as_dict,
            meta_class_getter=meta_class_getter,
            strict=False,
        )

        pretrained_name = pretrained_config.load_module_name
        loaded_and_renamed_meta_model = load_model(
            model_path=load_model_path,
            model_class=meta_model_class,
            model_init_kwargs=meta_model_kwargs,
            device="cpu",
            test_mode=False,
            state_dict_key_rename=replace_pattern,
            state_dict_keys_to_keep=(pretrained_name,),
        )
        loaded_and_renamed_fusion_extractors = (
            loaded_and_renamed_meta_model.input_modules
        )

        module_name_to_load = pretrained_config.load_module_name
        module_to_overload = loaded_and_renamed_fusion_extractors[input_name]

        logger.info(
            "Replacing '%s' in current model with '%s' from %s.",
            input_name,
            module_name_to_load,
            load_model_path,
        )

        input_modules[input_name] = module_to_overload

    return input_modules


def _build_all_replacements_tuples_for_loading_pretrained_module(
    input_configs: Sequence[schemas.InputConfig],
) -> Sequence[Tuple[str, str]]:
    replacement_patterns = []
    for input_config in input_configs:
        if input_config.pretrained_config:
            cur_replacement = _build_replace_tuple_when_loading_pretrained_module(
                load_module_name=input_config.pretrained_config.load_module_name,
                current_input_name=input_config.input_info.input_name,
            )
            if cur_replacement:
                replacement_patterns.append(cur_replacement)

    return replacement_patterns


def _build_replace_tuple_when_loading_pretrained_module(
    load_module_name: str, current_input_name: str
) -> Union[None, Tuple[str, str]]:
    if load_module_name == current_input_name:
        return None

    load_module_name_parsed = f"modules_to_fuse.{load_module_name}."
    current_input_name_parsed = f"modules_to_fuse.{current_input_name}."

    replace_pattern = (load_module_name_parsed, current_input_name_parsed)

    return replace_pattern



from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import torch
from aislib.misc_utils import ensure_path_exists
from PIL import Image
from torch.utils.data import Dataset
from torch.utils.data._utils.collate import default_collate

from eir.data_load.data_preparation_modules.imputation import (
    impute_missing_modalities_wrapper,
)
from eir.data_load.data_preparation_modules.prepare_array import un_normalize_array
from eir.data_load.datasets import al_getitem_return
from eir.interpretation.interpret_image import un_normalize_image
from eir.models.model_training_utils import predict_on_batch
from eir.setup.input_setup_modules.setup_array import ArrayNormalizationStats
from eir.setup.input_setup_modules.setup_image import ImageNormalizationStats
from eir.setup.output_setup_modules.array_output_setup import ComputedArrayOutputInfo
from eir.setup.output_setup_modules.image_output_setup import ComputedImageOutputInfo
from eir.setup.schemas import (
    ArrayOutputSamplingConfig,
    ArrayOutputTypeConfig,
    ImageOutputSamplingConfig,
    ImageOutputTypeConfig,
    OutputConfig,
)
from eir.train_utils import utils
from eir.train_utils.evaluation_handlers.evaluation_handlers_utils import (
    convert_model_inputs_to_raw,
    general_pre_process_prepared_inputs,
    get_batch_generator,
    get_dataset_loader_single_sample_generator,
    prepare_base_input,
    prepare_manual_sample_data,
    serialize_raw_inputs,
)
from eir.train_utils.step_modules.diffusion import p_sample_loop
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.predict import PredictExperiment, PredictHooks
    from eir.serve import ServeExperiment
    from eir.train import Experiment, Hooks, al_input_objects_as_dict

logger = get_logger(name=__name__)


@dataclass
class ArrayOutputEvalSamples:
    auto_samples: Dict[str, list["ArrayOutputEvalSample"]]
    manual_samples: Dict[str, list["ArrayOutputEvalSample"]]


@dataclass()
class ArrayOutputEvalSample:
    inputs_to_model: Dict[str, Any]
    target_labels: Dict[str, Any]
    sample_id: str


def array_out_single_sample_evaluation_wrapper(
    experiment: Union["Experiment", "PredictExperiment"],
    input_objects: "al_input_objects_as_dict",
    auto_dataset_to_load_from: Dataset,
    iteration: int,
    output_folder: str,
) -> None:
    default_eir_hooks = experiment.hooks

    output_configs = experiment.configs.output_configs

    if not any(i.sampling_config for i in output_configs):
        return

    manual_samples = get_array_output_manual_input_samples(
        output_configs=output_configs,
        input_objects=input_objects,
    )

    auto_validation_generator = get_dataset_loader_single_sample_generator(
        dataset=auto_dataset_to_load_from
    )
    auto_samples = get_array_output_auto_validation_samples(
        output_configs=output_configs,
        eval_sample_iterator=auto_validation_generator,
    )
    eval_samples_base = ArrayOutputEvalSamples(
        auto_samples=auto_samples,
        manual_samples=manual_samples,
    )

    for config in output_configs:
        if config.output_info.output_type not in ("array", "image"):
            continue

        cur_input_name = config.output_info.output_name
        cur_output_name = config.output_info.output_name

        not_in_manual_samples = cur_output_name not in eval_samples_base.manual_samples
        not_in_auto_samples = cur_output_name not in eval_samples_base.auto_samples
        if not_in_manual_samples and not_in_auto_samples:
            continue

        cur_sample_output_folder = utils.prepare_sample_output_folder(
            output_folder=output_folder,
            output_name=cur_output_name,
            column_name=cur_input_name,
            iteration=iteration,
        )

        output_type_info = config.output_type_info
        assert isinstance(
            output_type_info, (ArrayOutputTypeConfig, ImageOutputTypeConfig)
        )

        assert config.sampling_config is not None

        sample_generator = _get_eval_sample_generator(
            eval_samples=eval_samples_base, output_name=cur_output_name
        )

        batch_generator = get_batch_generator(
            iterator=enumerate(sample_generator),
            batch_size=experiment.configs.gc.be.batch_size,
        )

        for batch in batch_generator:
            batch_indices, batch_eval_data = zip(*batch)
            batch_eval_types, batch_eval_samples = zip(*batch_eval_data)

            if output_type_info.loss == "diffusion":
                time_steps = output_type_info.diffusion_time_steps
                assert time_steps is not None
                batch_generated_arrays = reverse_diffusion_array_generation(
                    eval_samples=batch_eval_samples,
                    array_output_name=cur_output_name,
                    experiment=experiment,
                    default_eir_hooks=default_eir_hooks,
                    num_steps=time_steps,
                )
            else:
                batch_generated_arrays = one_shot_array_generation(
                    eval_samples=batch_eval_samples,
                    array_output_name=cur_output_name,
                    experiment=experiment,
                    default_eir_hooks=default_eir_hooks,
                )

            for eval_type, idx, eval_sample, generated_array in zip(
                batch_eval_types,
                batch_indices,
                batch_eval_samples,
                batch_generated_arrays,
            ):

                cur_output_path = (
                    cur_sample_output_folder / eval_type / f"{idx}_generated.npy"
                )
                match output_type_info:
                    case ArrayOutputTypeConfig():
                        save_array_output(
                            array=generated_array, output_path=cur_output_path
                        )
                    case ImageOutputTypeConfig():
                        cur_output_path = cur_output_path.with_suffix(".png")
                        save_image_output(
                            array=generated_array, output_path=cur_output_path
                        )

                cur_inputs_output_path = (
                    cur_sample_output_folder / eval_type / f"{idx}_inputs"
                )

                raw_inputs = convert_model_inputs_to_raw(
                    inputs_to_model=eval_sample.inputs_to_model,
                    input_objects=input_objects,
                )

                serialize_raw_inputs(
                    raw_inputs=raw_inputs,
                    input_objects=input_objects,
                    output_path=cur_inputs_output_path,
                )


@torch.no_grad()
def one_shot_array_generation(
    eval_samples: tuple[ArrayOutputEvalSample, ...],
    array_output_name: str,
    experiment: Union["Experiment", "PredictExperiment", "ServeExperiment"],
    default_eir_hooks: Union["Hooks", "PredictHooks"],
) -> list[np.ndarray]:
    output_object = experiment.outputs[array_output_name]

    assert isinstance(output_object, (ComputedArrayOutputInfo, ComputedImageOutputInfo))

    array_sampling_batch = prepare_array_sampling_batch(
        eval_samples=eval_samples, array_output_name=array_output_name
    )
    prepared_sample_inputs = array_sampling_batch.prepared_inputs
    prepared_targets = array_sampling_batch.prepared_targets

    all_inputs = default_collate(prepared_sample_inputs)
    all_targets = default_collate(prepared_targets)
    all_ids = [eval_sample.sample_id for eval_sample in eval_samples]

    batch = general_pre_process_prepared_inputs(
        prepared_inputs=all_inputs,
        target_labels=all_targets,
        sample_ids=all_ids,
        experiment=experiment,
        custom_hooks=default_eir_hooks,
    )

    outputs = predict_on_batch(
        model=experiment.model,
        inputs=batch.inputs,
    )

    batch_size = len(eval_samples)
    array_outputs = outputs[array_output_name][array_output_name]

    assert output_object.normalization_stats is not None

    final_numpy_outputs = []
    for batch_idx in range(batch_size):
        cur_output = array_outputs[batch_idx]
        cur_output_raw = un_normalize_wrapper(
            array=cur_output,
            normalization_stats=output_object.normalization_stats,
        )
        cur_output_numpy = cur_output_raw.cpu().numpy()
        final_numpy_outputs.append(cur_output_numpy)

    return final_numpy_outputs


@torch.inference_mode()
def reverse_diffusion_array_generation(
    eval_samples: tuple[ArrayOutputEvalSample, ...],
    array_output_name: str,
    experiment: Union["Experiment", "PredictExperiment", "ServeExperiment"],
    default_eir_hooks: Union["Hooks", "PredictHooks"],
    num_steps: int,
) -> list[np.ndarray]:
    """
    TODO: Consider not keeping / making configurable whether to keep intermediate
          states as this might blow up memory usage in some cases.
    """
    output_object = experiment.outputs[array_output_name]
    assert isinstance(output_object, (ComputedArrayOutputInfo, ComputedImageOutputInfo))

    dimensions = output_object.data_dimensions
    batch_size = len(eval_samples)
    shape = (batch_size,) + dimensions.full_shape()

    array_sampling_batch = prepare_array_sampling_batch(
        eval_samples=eval_samples,
        array_output_name=array_output_name,
    )
    prepared_sample_inputs = array_sampling_batch.prepared_inputs
    prepared_targets = array_sampling_batch.prepared_targets

    all_inputs = default_collate(prepared_sample_inputs)
    all_targets = default_collate(prepared_targets)
    all_ids = [eval_sample.sample_id for eval_sample in eval_samples]

    batch = general_pre_process_prepared_inputs(
        prepared_inputs=all_inputs,
        target_labels=all_targets,
        sample_ids=all_ids,
        experiment=experiment,
        custom_hooks=default_eir_hooks,
    )
    batch_inputs = deepcopy(batch.inputs)

    dc = output_object.diffusion_config
    assert dc is not None

    final_states = p_sample_loop(
        config=dc,
        batch_inputs=batch_inputs,
        output_name=array_output_name,
        model=experiment.model,
        output_shape=shape,
        time_steps=num_steps,
    )

    assert output_object.normalization_stats is not None
    final_numpy_outputs = []
    for batch_idx in range(batch_size):
        cur_sample_final_state = final_states[batch_idx]
        cur_final_output = un_normalize_wrapper(
            array=torch.from_numpy(cur_sample_final_state),
            normalization_stats=output_object.normalization_stats,
        )
        cur_final_output_numpy = cur_final_output.cpu().numpy()
        final_numpy_outputs.append(cur_final_output_numpy)

    return final_numpy_outputs


@dataclass(frozen=False)
class ArraySamplingBatch:
    prepared_inputs: list[dict[str, torch.Tensor]]
    prepared_targets: list[dict[str, torch.Tensor]]


def prepare_array_sampling_batch(
    eval_samples: Sequence[ArrayOutputEvalSample],
    array_output_name: str,
) -> ArraySamplingBatch:
    """
    Every sample has an extra batch dimension by default from e.g. the data
    loader and preparation. We squeeze here as we collate multiple samples
    later to avoid an extra dimension.
    """
    prepared_sample_inputs = []
    prepared_targets = []

    for eval_sample in eval_samples:
        cur_inputs = eval_sample.inputs_to_model
        cur_prepared = prepare_base_input(prepared_inputs=cur_inputs)

        prepared_sample_inputs.append(cur_prepared)
        prepared_targets.append(eval_sample.target_labels)

    return ArraySamplingBatch(
        prepared_inputs=prepared_sample_inputs,
        prepared_targets=prepared_targets,
    )


def un_normalize_wrapper(
    array: torch.Tensor,
    normalization_stats: ArrayNormalizationStats | ImageNormalizationStats,
) -> torch.Tensor:

    match normalization_stats:
        case ArrayNormalizationStats():
            un_normalized = un_normalize_array(
                array=array,
                normalization_stats=normalization_stats,
            )
        case ImageNormalizationStats():
            un_normalized_arr = un_normalize_image(
                normalized_img=array.cpu().numpy(),
                normalization_stats=normalization_stats,
            )
            un_normalized = torch.from_numpy(un_normalized_arr)
        case _:
            raise ValueError("Normalization stats not recognized.")

    return un_normalized


def save_array_output(array: np.ndarray, output_path: Path | str) -> None:
    ensure_path_exists(path=Path(output_path))
    np.save(file=output_path, arr=array)


def save_image_output(array: np.ndarray, output_path: Path | str) -> None:
    ensure_path_exists(path=Path(output_path))

    assert array.ndim == 3, "Input should be 3D"

    array_hwc = np.moveaxis(array, 0, -1)
    array_uint8 = (array_hwc * 255).astype(np.uint8)

    n_channels = array_uint8.shape[-1]
    mode: Optional[str]
    match n_channels:
        case 1:
            mode = "L"
            array_uint8 = array_uint8.squeeze(-1)
        case 3:
            mode = "RGB"
        case 4:
            mode = "RGBA"
        case _:
            mode = None

    pil_image = Image.fromarray(array_uint8, mode=mode)
    pil_image.save(output_path)


def get_array_output_manual_input_samples(
    output_configs: Sequence[OutputConfig],
    input_objects: "al_input_objects_as_dict",
) -> Dict[str, list[ArrayOutputEvalSample]]:
    prepared_samples: dict[str, list[ArrayOutputEvalSample]] = {}

    for config_idx, config in enumerate(output_configs):
        array_or_image = config.output_info.output_type in ("array", "image")
        if not config.sampling_config or not array_or_image:
            continue

        assert isinstance(
            config.sampling_config,
            (ArrayOutputSamplingConfig, ImageOutputSamplingConfig),
        )

        sample_data_from_yaml = config.sampling_config.manual_inputs
        output_name = config.output_info.output_name

        prepared_samples[output_name] = []

        for idx, single_sample_inputs in enumerate(sample_data_from_yaml):
            prepared_inputs = prepare_manual_sample_data(
                sample_inputs=single_sample_inputs,
                input_objects=input_objects,
            )

            imputed_inputs = impute_missing_modalities_wrapper(
                inputs_values=prepared_inputs,
                inputs_objects=input_objects,
            )

            cur_eval_sample = ArrayOutputEvalSample(
                inputs_to_model=imputed_inputs,
                target_labels={},
                sample_id=f"manual_{idx}",
            )

            prepared_samples[output_name].append(cur_eval_sample)

    return prepared_samples


def get_array_output_auto_validation_samples(
    output_configs: Sequence[OutputConfig],
    eval_sample_iterator: Iterator[al_getitem_return],
) -> Dict[str, list[ArrayOutputEvalSample]]:
    prepared_eval_samples: dict[str, list[ArrayOutputEvalSample]] = {}

    for config_idx, config in enumerate(output_configs):
        array_or_image = config.output_info.output_type in ("array", "image")
        if not config.sampling_config or not array_or_image:
            continue

        output_name = config.output_info.output_name

        prepared_eval_samples[output_name] = []

        assert isinstance(
            config.sampling_config,
            (ArrayOutputSamplingConfig, ImageOutputSamplingConfig),
        )
        for i in range(config.sampling_config.n_eval_inputs):
            input_to_model, target_labels, cur_id = next(eval_sample_iterator)

            cur_eval_sample = ArrayOutputEvalSample(
                inputs_to_model=input_to_model,
                target_labels=target_labels,
                sample_id=cur_id,
            )

            prepared_eval_samples[output_name].append(cur_eval_sample)

    return prepared_eval_samples


def _get_eval_sample_generator(
    eval_samples: ArrayOutputEvalSamples, output_name: str
) -> Generator[Tuple[str, ArrayOutputEvalSample], None, None]:
    cur_config_auto_samples = eval_samples.auto_samples[output_name]
    cur_config_manual_samples = eval_samples.manual_samples[output_name]

    for eval_sample in cur_config_auto_samples:
        yield "auto", eval_sample

    for eval_sample in cur_config_manual_samples:
        yield "manual", eval_sample



from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Optional

import numpy as np
import torch

from eir.data_load.data_source_modules.deeplake_ops import (
    get_deeplake_input_source_iterable,
    is_deeplake_dataset,
    load_deeplake_dataset,
)
from eir.setup import schemas
from eir.setup.input_setup_modules.common import (
    DataDimensions,
    get_data_dimension_from_data_source,
    get_dtype_from_data_source,
)
from eir.setup.setup_utils import (
    ChannelBasedRunningStatistics,
    ElementBasedRunningStatistics,
    al_collector_classes,
    collect_stats,
)
from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


@dataclass
class ComputedArrayInputInfo:
    input_config: schemas.InputConfig
    data_dimensions: "DataDimensions"
    dtype: np.dtype
    normalization_stats: Optional["ArrayNormalizationStats"] = None


def set_up_array_input(
    input_config: schemas.InputConfig, *args, **kwargs
) -> ComputedArrayInputInfo:
    data_dimensions = get_data_dimension_from_data_source(
        data_source=Path(input_config.input_info.input_source),
        deeplake_inner_key=input_config.input_info.input_inner_key,
    )

    input_type_info = input_config.input_type_info
    assert isinstance(input_type_info, schemas.ArrayInputDataConfig)

    normalization_stats: Optional[ArrayNormalizationStats] = None
    if input_type_info.normalization is not None:
        normalization_stats = get_array_normalization_values(
            source=input_config.input_info.input_source,
            inner_key=input_config.input_info.input_inner_key,
            normalization=input_type_info.normalization,
            data_dimensions=data_dimensions,
            max_samples=input_type_info.adaptive_normalization_max_samples,
        )

    dtype = get_dtype_from_data_source(
        data_source=Path(input_config.input_info.input_source),
        deeplake_inner_key=input_config.input_info.input_inner_key,
    )

    array_input_info = ComputedArrayInputInfo(
        input_config=input_config,
        data_dimensions=data_dimensions,
        dtype=dtype,
        normalization_stats=normalization_stats,
    )

    return array_input_info


@dataclass
class ArrayNormalizationStats:
    shape: tuple
    means: torch.Tensor
    stds: torch.Tensor
    type: Optional[Literal["element", "channel"]]


def get_array_normalization_values(
    source: str,
    inner_key: Optional[str],
    normalization: Optional[Literal["element", "channel"]],
    data_dimensions: DataDimensions,
    max_samples: Optional[int],
) -> ArrayNormalizationStats:
    input_source = source
    deeplake_inner_key = inner_key

    if is_deeplake_dataset(data_source=input_source):
        deeplake_ds = load_deeplake_dataset(data_source=input_source)
        assert deeplake_inner_key is not None
        ds_iter = get_deeplake_input_source_iterable(
            deeplake_dataset=deeplake_ds, inner_key=deeplake_inner_key
        )
        tensor_iterator = (torch.from_numpy(i.numpy()).float() for i in ds_iter)
    else:
        file_iterator = Path(input_source).rglob("*")
        np_iterator = (np.load(str(i)) for i in file_iterator)
        tensor_iterator = (torch.from_numpy(i).float() for i in np_iterator)

    tensor_iterator = (i.reshape(data_dimensions.full_shape()) for i in tensor_iterator)

    collector_class: al_collector_classes
    if normalization == "channel":
        collector_class = ChannelBasedRunningStatistics
    elif normalization == "element":
        collector_class = ElementBasedRunningStatistics
    else:
        raise ValueError(
            f"Invalid normalization type: {normalization}. "
            f"Must be one of ['element', 'channel']"
        )

    gathered_stats = collect_stats(
        tensor_iterable=tensor_iterator,
        collector_class=collector_class,
        shape=data_dimensions.full_shape(),
        max_samples=max_samples,
        name=source,
    )

    means = _add_extra_dims_if_needed(
        normalization_tensor=gathered_stats.mean,
        data_shape=data_dimensions.full_shape(),
        normalization_type=normalization,
    )

    stds = _add_extra_dims_if_needed(
        normalization_tensor=gathered_stats.std,
        data_shape=data_dimensions.full_shape(),
        normalization_type=normalization,
    )

    normalization_stats = ArrayNormalizationStats(
        shape=data_dimensions.full_shape(),
        means=means,
        stds=stds,
        type=normalization,
    )

    _check_normalization_stats(source=source, normalization_stats=normalization_stats)

    return normalization_stats


def _add_extra_dims_if_needed(
    normalization_tensor: torch.Tensor,
    data_shape: tuple,
    normalization_type: Optional[Literal["element", "channel"]],
) -> torch.Tensor:
    if normalization_type == "element":
        assert normalization_tensor.shape == data_shape
        return normalization_tensor

    elif normalization_type == "channel":
        assert normalization_tensor.dim() == 1
        assert normalization_tensor.shape[0] == data_shape[0]

        num_extra_dims = len(data_shape) - 1
        new_shape = (data_shape[0],) + (1,) * num_extra_dims
        reshaped_tensor = normalization_tensor.view(*new_shape)

        return reshaped_tensor

    else:
        raise ValueError(
            f"Invalid normalization type: {normalization_type}. "
            f"Must be one of ['element', 'channel']"
        )


def _check_normalization_stats(
    source: str, normalization_stats: ArrayNormalizationStats, epsilon: float = 1e-10
) -> None:
    if torch.any(normalization_stats.stds < epsilon):
        num_zero_stds = torch.sum(normalization_stats.stds < epsilon).item()
        logger.warning(
            f"In source {source}, "
            f"{num_zero_stds} elements have zero or near-zero standard deviation. "
            "This may lead to large values when normalizing. "
            "Consider handling these cases specifically.",
        )
